# type:ignore

"""Test chat model integration."""

import json
import os
from contextlib import nullcontext
from typing import Any, Callable, Dict, Literal, Type, cast
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import AIMessage, HumanMessage, SystemMessage, ToolMessage
from langchain_core.runnables import RunnableBinding
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field

from langchain_aws import ChatBedrock
from langchain_aws.chat_models.bedrock import (
    ChatPromptAdapter,
    _format_anthropic_messages,
    _merge_messages,
    convert_messages_to_prompt_anthropic,
)
from langchain_aws.function_calling import convert_to_anthropic_tool


def test_profile() -> None:
    model = ChatBedrock(
        model_id="anthropic.claude-3-5-sonnet-20241022-v2:0",
        region_name="us-west-2",
    )
    assert model.profile
    assert not model.profile["reasoning_output"]

    model = ChatBedrock(
        model_id="anthropic.claude-sonnet-4-20250514-v1:0",
        region_name="us-west-2",
    )
    assert model.profile
    assert model.profile["reasoning_output"]

    model = ChatBedrock(model_id="foo")
    assert model.profile == {}


def test__merge_messages() -> None:
    messages = [
        SystemMessage("foo"),  # type: ignore[misc]
        SystemMessage("barfoo"),  # type: ignore[misc]
        HumanMessage("bar"),  # type: ignore[misc]
        AIMessage(  # type: ignore[misc]
            [
                {"text": "baz", "type": "text"},
                {
                    "tool_input": {"a": "b"},
                    "type": "tool_use",
                    "id": "1",
                    "text": None,
                    "name": "buz",
                },
                {"text": "baz", "type": "text"},
                {
                    "tool_input": {"a": "c"},
                    "type": "tool_use",
                    "id": "2",
                    "text": None,
                    "name": "blah",
                },
            ]
        ),
        ToolMessage("buz output", tool_call_id="1"),  # type: ignore[misc]
        ToolMessage("blah output", tool_call_id="2"),  # type: ignore[misc]
        HumanMessage("next thing"),  # type: ignore[misc]
    ]
    expected = [
        SystemMessage(
            [{"type": "text", "text": "foo"}, {"type": "text", "text": "barfoo"}]
        ),  # type: ignore[misc]
        HumanMessage("bar"),  # type: ignore[misc]
        AIMessage(  # type: ignore[misc]
            [
                {"text": "baz", "type": "text"},
                {
                    "tool_input": {"a": "b"},
                    "type": "tool_use",
                    "id": "1",
                    "text": None,
                    "name": "buz",
                },
                {"text": "baz", "type": "text"},
                {
                    "tool_input": {"a": "c"},
                    "type": "tool_use",
                    "id": "2",
                    "text": None,
                    "name": "blah",
                },
            ]
        ),
        HumanMessage(  # type: ignore[misc]
            [
                {"type": "tool_result", "content": "buz output", "tool_use_id": "1"},
                {"type": "tool_result", "content": "blah output", "tool_use_id": "2"},
                {"type": "text", "text": "next thing"},
            ]
        ),
    ]
    actual = _merge_messages(messages)
    assert expected == actual


def test__merge_messages_mutation() -> None:
    original_messages = [
        HumanMessage([{"type": "text", "text": "bar"}]),  # type: ignore[misc]
        HumanMessage("next thing"),  # type: ignore[misc]
    ]
    messages = [
        HumanMessage([{"type": "text", "text": "bar"}]),  # type: ignore[misc]
        HumanMessage("next thing"),  # type: ignore[misc]
    ]
    expected = [
        HumanMessage(  # type: ignore[misc]
            [{"type": "text", "text": "bar"}, {"type": "text", "text": "next thing"}]
        ),
    ]
    actual = _merge_messages(messages)
    assert expected == actual
    assert messages == original_messages


def test__format_anthropic_messages_with_tool_calls() -> None:
    system = SystemMessage("fuzz")  # type: ignore[misc]
    human = HumanMessage("foo")  # type: ignore[misc]
    ai = AIMessage(  # type: ignore[misc]
        "",
        tool_calls=[{"name": "bar", "id": "1", "args": {"baz": "buzz"}}],
    )
    tool = ToolMessage(  # type: ignore[misc]
        "blurb",
        tool_call_id="1",
    )
    human_2 = HumanMessage("try again.")
    messages = [system, human, ai, tool, human_2]
    expected = (
        "fuzz",
        [
            {"role": "user", "content": "foo"},
            {
                "role": "assistant",
                "content": [
                    {
                        "type": "tool_use",
                        "name": "bar",
                        "id": "1",
                        "input": {"baz": "buzz"},
                    }
                ],
            },
            {
                "role": "user",
                "content": [
                    {"type": "tool_result", "content": "blurb", "tool_use_id": "1"},
                    {"type": "text", "text": "try again."},
                ],
            },
        ],
    )
    actual = _format_anthropic_messages(messages)
    assert expected == actual


def test__format_anthropic_messages_with_str_content_and_tool_calls() -> None:
    system = SystemMessage("fuzz")  # type: ignore[misc]
    human = HumanMessage("foo")  # type: ignore[misc]
    # If content and tool_calls are specified and content is a string, then both are
    # included with content first.
    ai = AIMessage(  # type: ignore[misc]
        "thought",
        tool_calls=[{"name": "bar", "id": "1", "args": {"baz": "buzz"}}],
    )
    tool = ToolMessage("blurb", tool_call_id="1")  # type: ignore[misc]
    messages = [system, human, ai, tool]
    expected = (
        "fuzz",
        [
            {"role": "user", "content": "foo"},
            {
                "role": "assistant",
                "content": [
                    {"type": "text", "text": "thought"},
                    {
                        "type": "tool_use",
                        "name": "bar",
                        "id": "1",
                        "input": {"baz": "buzz"},
                    },
                ],
            },
            {
                "role": "user",
                "content": [
                    {"type": "tool_result", "content": "blurb", "tool_use_id": "1"}
                ],
            },
        ],
    )
    actual = _format_anthropic_messages(messages)
    assert expected == actual


def test__format_anthropic_messages_with_list_content_and_tool_calls() -> None:
    system = SystemMessage("fuzz")  # type: ignore[misc]
    human = HumanMessage("foo")  # type: ignore[misc]
    # If content and tool_calls are specified and content is a list, both are
    # included
    ai = AIMessage(  # type: ignore[misc]
        [{"type": "text", "text": "thought"}],
        tool_calls=[{"name": "bar", "id": "1", "args": {"baz": "buzz"}}],
    )
    tool = ToolMessage(  # type: ignore[misc]
        "blurb",
        tool_call_id="1",
    )
    messages = [system, human, ai, tool]
    expected = (
        "fuzz",
        [
            {"role": "user", "content": "foo"},
            {
                "role": "assistant",
                "content": [
                    {"type": "text", "text": "thought"},
                    {
                        "type": "tool_use",
                        "name": "bar",
                        "id": "1",
                        "input": {"baz": "buzz"},
                    },
                ],
            },
            {
                "role": "user",
                "content": [
                    {"type": "tool_result", "content": "blurb", "tool_use_id": "1"}
                ],
            },
        ],
    )
    actual = _format_anthropic_messages(messages)
    assert expected == actual


def test__format_anthropic_messages_with_tool_use_blocks_and_tool_calls() -> None:
    """Show that tool_calls are preferred to tool_use blocks when both have same id."""
    system = SystemMessage("fuzz")  # type: ignore[misc]
    human = HumanMessage("foo")  # type: ignore[misc]
    # NOTE: tool_use block in contents and tool_calls have different arguments.
    ai = AIMessage(  # type: ignore[misc]
        [
            {"type": "text", "text": "thought"},
            {
                "type": "tool_use",
                "name": "bar",
                "id": "1",
                "input": {"baz": "NOT_BUZZ"},
            },
        ],
        tool_calls=[{"name": "bar", "id": "1", "args": {"baz": "BUZZ"}}],
    )
    tool = ToolMessage("blurb", tool_call_id="1")  # type: ignore[misc]
    messages = [system, human, ai, tool]
    expected = (
        "fuzz",
        [
            {"role": "user", "content": "foo"},
            {
                "role": "assistant",
                "content": [
                    {"type": "text", "text": "thought"},
                    {
                        "type": "tool_use",
                        "name": "bar",
                        "id": "1",
                        "input": {"baz": "BUZZ"},  # tool_calls value preferred.
                    },
                ],
            },
            {
                "role": "user",
                "content": [
                    {"type": "tool_result", "content": "blurb", "tool_use_id": "1"}
                ],
            },
        ],
    )
    actual = _format_anthropic_messages(messages)
    assert expected == actual


def test__format_anthropic_messages_with_cache_control() -> None:
    system = SystemMessage(
        [
            {
                "type": "text",
                "text": "fuzz",
                "cache_control": {"type": "ephemeral"},
            },
            "bar",
        ],
    )
    human = HumanMessage(
        [
            {
                "type": "text",
                "text": "foo",
                "cache_control": {"type": "ephemeral"},
            },
        ],
    )
    messages = [system, human]
    expected = (
        [
            {
                "type": "text",
                "text": "fuzz",
                "cache_control": {"type": "ephemeral"},
            },
            {"type": "text", "text": "bar"},
        ],
        [
            {
                "role": "user",
                "content": [
                    {
                        "type": "text",
                        "text": "foo",
                        "cache_control": {"type": "ephemeral"},
                    }
                ],
            },
        ],
    )

    actual = _format_anthropic_messages(messages)
    assert expected == actual


def test__format_anthropic_messages_system_message_list_content() -> None:
    """Test that system messages with list content return as list of content blocks"""
    system = SystemMessage(
        [
            {"type": "text", "text": "You are a helpful assistant."},
            {"type": "text", "text": "Additional instructions here."},
        ],
    )
    human = HumanMessage("Hello!")
    messages = [system, human]
    expected = (
        [
            {"type": "text", "text": "You are a helpful assistant."},
            {"type": "text", "text": "Additional instructions here."},
        ],
        [
            {"role": "user", "content": "Hello!"},
        ],
    )

    actual = _format_anthropic_messages(messages)
    assert expected == actual


def test__format_anthropic_multiple_system_messages() -> None:
    """Test that multiple system messages can be passed, and that none of them are required to be at position 0."""  # noqa: E501
    system1 = SystemMessage("foo")  # type: ignore[misc]
    system2 = SystemMessage("bar")  # type: ignore[misc]
    human = HumanMessage("Hello!")
    messages = [human, system1, system2]
    expected_system = [{"text": "foo", "type": "text"}, {"text": "bar", "type": "text"}]
    expected_messages = [{"role": "user", "content": "Hello!"}]

    actual_system, actual_messages = _format_anthropic_messages(messages)
    assert expected_system == actual_system
    assert expected_messages == actual_messages


def test__format_anthropic_nonconsecutive_system_messages() -> None:
    """Test that we fail when non-consecutive system messages are passed."""
    system1 = SystemMessage("foo")  # type: ignore[misc]
    system2 = SystemMessage("bar")  # type: ignore[misc]
    human = HumanMessage("Hello!")
    messages = [system1, human, system2]

    with pytest.raises(
        ValueError, match="Received multiple non-consecutive system messages."
    ):
        _format_anthropic_messages(messages)


@pytest.fixture()
def pydantic() -> Type[BaseModel]:
    class dummy_function(BaseModel):
        """dummy function"""

        arg1: int = Field(..., description="foo")
        arg2: Literal["bar", "baz"] = Field(..., description="one of 'bar', 'baz'")

    return dummy_function


@pytest.fixture()
def function() -> Callable:
    def dummy_function(arg1: int, arg2: Literal["bar", "baz"]) -> None:
        """dummy function

        Args:
            arg1: foo
            arg2: one of 'bar', 'baz'
        """
        pass

    return dummy_function


@pytest.fixture()
def dummy_tool() -> BaseTool:
    class Schema(BaseModel):
        arg1: int = Field(..., description="foo")
        arg2: Literal["bar", "baz"] = Field(..., description="one of 'bar', 'baz'")

    class DummyFunction(BaseTool):
        args_schema: Type[BaseModel] = Schema
        name: str = "dummy_function"
        description: str = "dummy function"

        def _run(self, *args: Any, **kwargs: Any) -> Any:
            pass

    return DummyFunction()


@pytest.fixture()
def json_schema() -> Dict:
    return {
        "title": "dummy_function",
        "description": "dummy function",
        "type": "object",
        "properties": {
            "arg1": {"description": "foo", "type": "integer"},
            "arg2": {
                "description": "one of 'bar', 'baz'",
                "enum": ["bar", "baz"],
                "type": "string",
            },
        },
        "required": ["arg1", "arg2"],
    }


@pytest.fixture()
def openai_function() -> Dict:
    return {
        "name": "dummy_function",
        "description": "dummy function",
        "parameters": {
            "type": "object",
            "properties": {
                "arg1": {"description": "foo", "type": "integer"},
                "arg2": {
                    "description": "one of 'bar', 'baz'",
                    "enum": ["bar", "baz"],
                    "type": "string",
                },
            },
            "required": ["arg1", "arg2"],
        },
    }


@pytest.fixture()
def tool_with_empty_description() -> Dict:
    return {
        "name": "dummy_function",
        "parameters": {
            "type": "object",
            "properties": {
                "arg1": {"description": "foo", "type": "integer"},
                "arg2": {
                    "description": "one of 'bar', 'baz'",
                    "enum": ["bar", "baz"],
                    "type": "string",
                },
            },
            "required": ["arg1", "arg2"],
        },
    }


def test_convert_to_anthropic_tool(
    pydantic: Type[BaseModel],
    function: Callable,
    dummy_tool: BaseTool,
    json_schema: Dict,
    openai_function: Dict,
    tool_with_empty_description: Dict,
) -> None:
    expected = {
        "name": "dummy_function",
        "description": "dummy function",
        "input_schema": {
            "type": "object",
            "properties": {
                "arg1": {"description": "foo", "type": "integer"},
                "arg2": {
                    "description": "one of 'bar', 'baz'",
                    "enum": ["bar", "baz"],
                    "type": "string",
                },
            },
            "required": ["arg1", "arg2"],
        },
    }

    for fn in (pydantic, function, dummy_tool, json_schema, expected, openai_function):
        actual = convert_to_anthropic_tool(fn)  # type: ignore[arg-type]
        assert actual == expected

    expected["description"] = expected["name"]
    actual = convert_to_anthropic_tool(tool_with_empty_description)
    assert actual == expected


class GetWeather(BaseModel):
    """Get the current weather in a given location"""

    location: str = Field(..., description="The city and state, e.g. San Francisco, CA")


def test_anthropic_bind_tools_tool_choice() -> None:
    chat_model = ChatBedrock(
        model_id="anthropic.claude-opus-4-1-20250805-v1:0", region_name="us-west-2"
    )  # type: ignore[call-arg]
    chat_model_with_tools = chat_model.bind_tools(
        [GetWeather], tool_choice={"type": "tool", "name": "GetWeather"}
    )
    assert cast(RunnableBinding, chat_model_with_tools).kwargs["tool_choice"] == {
        "type": "tool",
        "name": "GetWeather",
    }
    chat_model_with_tools = chat_model.bind_tools(
        [GetWeather], tool_choice="GetWeather"
    )
    assert cast(RunnableBinding, chat_model_with_tools).kwargs["tool_choice"] == {
        "type": "tool",
        "name": "GetWeather",
    }
    chat_model_with_tools = chat_model.bind_tools([GetWeather], tool_choice="auto")
    assert cast(RunnableBinding, chat_model_with_tools).kwargs["tool_choice"] == {
        "type": "auto"
    }
    chat_model_with_tools = chat_model.bind_tools([GetWeather], tool_choice="any")
    assert cast(RunnableBinding, chat_model_with_tools).kwargs["tool_choice"] == {
        "type": "any"
    }


@pytest.mark.parametrize(
    "model_id",
    [
        "anthropic.claude-sonnet-4-20250514-v1:0",
        "anthropic.claude-sonnet-4-5-20250929-v1:0",
        "anthropic.claude-haiku-4-5-20251001-v1:0",
    ],
)
@pytest.mark.parametrize(
    "tool_choice",
    [
        True,
        "any",
        "GetWeather",
        {"type": "tool", "name": "GetWeather"},
    ],
)
@mock.patch("langchain_aws.chat_models.bedrock.create_aws_client")
def test_claude_thinking_forced_tool_raises(
    mock_create_aws_client, model_id, tool_choice
) -> None:
    mock_create_aws_client.return_value = MagicMock()
    chat = ChatBedrock(
        model_id=model_id,
        region_name="us-west-2",
        model_kwargs={
            "thinking": {"type": "enabled", "budget_tokens": 2048},
        },
    )
    with pytest.raises(ValueError, match="does not support forced tool use"):
        chat.bind_tools([GetWeather], tool_choice=tool_choice)


@mock.patch("langchain_aws.chat_models.bedrock.create_aws_client")
def test_claude_thinking_tool_choice_auto_ok(mock_create_aws_client) -> None:
    mock_create_aws_client.return_value = MagicMock()
    chat = ChatBedrock(
        model_id="anthropic.claude-sonnet-4-5-20250929-v1:0",
        region_name="us-west-2",
        model_kwargs={
            "thinking": {"type": "enabled", "budget_tokens": 2048},
        },
    )
    chat_with_tools = chat.bind_tools([GetWeather], tool_choice="auto")
    assert cast(RunnableBinding, chat_with_tools).kwargs["tool_choice"] == {
        "type": "auto"
    }


@mock.patch("langchain_aws.chat_models.bedrock.create_aws_client")
def test_claude_thinking_with_structured_output_ok(mock_create_aws_client) -> None:
    """Test with_structured_output with thinking mode omits forced tool choice."""
    from pydantic import BaseModel

    class WeatherResponse(BaseModel):
        temperature: int
        condition: str

    mock_client = MagicMock()
    mock_create_aws_client.return_value = mock_client

    chat = ChatBedrock(
        model_id="anthropic.claude-sonnet-4-5-20250929-v1:0",
        region_name="us-west-2",
        model_kwargs={
            "thinking": {"type": "enabled", "budget_tokens": 2048},
        },
    )

    structured_llm = chat.with_structured_output(WeatherResponse)
    assert structured_llm is not None


@pytest.mark.parametrize(
    "model_id",
    [
        "anthropic.claude-sonnet-4-20250514-v1:0",
        "anthropic.claude-sonnet-4-5-20250929-v1:0",
        "anthropic.claude-haiku-4-5-20251001-v1:0",
    ],
)
@mock.patch("langchain_aws.chat_models.bedrock.create_aws_client")
def test_claude_models_no_thinking_forced_tool_ok(
    mock_create_aws_client, model_id
) -> None:
    mock_create_aws_client.return_value = MagicMock()
    chat = ChatBedrock(
        model_id=model_id,
        region_name="us-west-2",
    )
    chat_with_tools = chat.bind_tools([GetWeather], tool_choice="any")
    assert cast(RunnableBinding, chat_with_tools).kwargs["tool_choice"] == {
        "type": "any"
    }


@mock.patch("langchain_aws.chat_models.bedrock.create_aws_client")
def test_other_anthropic_model_thinking_forced_tool_ok(mock_create_aws_client) -> None:
    mock_create_aws_client.return_value = MagicMock()
    chat = ChatBedrock(
        model_id="anthropic.claude-3-5-sonnet-20241022-v2:0",
        region_name="us-west-2",
        model_kwargs={
            "thinking": {"type": "enabled", "budget_tokens": 2048},
        },
    )
    chat_with_tools = chat.bind_tools([GetWeather], tool_choice="any")
    assert cast(RunnableBinding, chat_with_tools).kwargs["tool_choice"] == {
        "type": "any"
    }


def test_bind_tools_strict_warns_without_converse_api() -> None:
    chat_model = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0", region="us-west-2"
    )  # type: ignore[call-arg]
    with pytest.warns(UserWarning, match="only supported when using the Converse API"):
        chat_model.bind_tools([GetWeather], strict=True)


@mock.patch("langchain_aws.chat_models.bedrock.create_aws_client")
def test_bind_tools_strict_passed_to_converse_api(
    mock_create_aws_client: Any,
) -> None:
    mock_create_aws_client.return_value = MagicMock()
    chat_model = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        region="us-west-2",
        beta_use_converse_api=True,
    )
    chat_model_with_tools = chat_model.bind_tools([GetWeather], strict=True)

    bound_kwargs = cast(RunnableBinding, chat_model_with_tools).kwargs
    func = bound_kwargs["tools"][0]["function"]
    assert func["strict"] is True


def test_with_structured_output_strict_warns_without_converse_api() -> None:
    chat_model = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0", region="us-west-2"
    )  # type: ignore[call-arg]
    with pytest.warns(UserWarning, match="only supported when using the Converse API"):
        chat_model.with_structured_output(GetWeather, strict=True)


@mock.patch("langchain_aws.chat_models.bedrock.create_aws_client")
def test_with_structured_output_strict_passed_to_converse_api(
    mock_create_aws_client: Any,
) -> None:
    mock_create_aws_client.return_value = MagicMock()
    chat_model = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        region="us-west-2",
        beta_use_converse_api=True,
    )
    structured = chat_model.with_structured_output(GetWeather, strict=True)
    llm = structured.first  # type: ignore[attr-defined]
    bound_kwargs = cast(RunnableBinding, llm).kwargs
    func = bound_kwargs["tools"][0]["function"]
    assert func["strict"] is True


def test_standard_tracing_params() -> None:
    llm = ChatBedrock(model_id="foo", region_name="us-west-2")  # type: ignore[call-arg]
    expected = {
        "ls_provider": "amazon_bedrock",
        "ls_model_type": "chat",
        "ls_model_name": "foo",
    }
    assert llm._get_ls_params() == expected

    # Test initialization with `model` alias
    llm = ChatBedrock(model="foo", region_name="us-west-2")
    assert llm._get_ls_params() == expected

    llm = ChatBedrock(
        model_id="foo", model_kwargs={"temperature": 0.1}, region_name="us-west-2"
    )  # type: ignore[call-arg]
    ls_params = llm._get_ls_params()
    assert ls_params == {
        "ls_provider": "amazon_bedrock",
        "ls_model_type": "chat",
        "ls_model_name": "foo",
        "ls_temperature": 0.1,
    }


def test_beta_use_converse_api() -> None:
    llm = ChatBedrock(model_id="amazon.nova.foo", region_name="us-west-2")  # type: ignore[call-arg]
    assert llm.beta_use_converse_api

    llm = ChatBedrock(
        model="foobar", base_model="amazon.nova.foo", region_name="us-west-2"
    )  # type: ignore[call-arg]
    assert llm.beta_use_converse_api

    llm = ChatBedrock(
        model="arn:aws:bedrock:::application-inference-profile/my-profile",
        base_model="claude.foo",
        region_name="us-west-2",
    )  # type: ignore[call-arg]
    assert not llm.beta_use_converse_api

    llm = ChatBedrock(
        model="nova.foo", region_name="us-west-2", beta_use_converse_api=False
    )
    assert not llm.beta_use_converse_api

    llm = ChatBedrock(
        model="foobar",
        base_model="nova.foo",
        region_name="us-west-2",
        beta_use_converse_api=False,
    )
    assert not llm.beta_use_converse_api

    llm = ChatBedrock(model="foo", region_name="us-west-2", beta_use_converse_api=True)
    assert llm.beta_use_converse_api

    llm = ChatBedrock(model="foo", region_name="us-west-2", beta_use_converse_api=False)
    assert not llm.beta_use_converse_api


@mock.patch("langchain_aws.chat_models.bedrock.create_aws_client")
def test_beta_use_converse_api_with_inference_profile(mock_create_aws_client):
    mock_bedrock_client = mock.MagicMock()
    mock_bedrock_client.get_inference_profile.return_value = {
        "models": [
            {
                "modelArn": "arn:aws:bedrock:us-west-2::foundation-model/anthropic.claude-3-sonnet-20240229-v1:0"  # noqa: E501
            }
        ]
    }
    mock_create_aws_client.return_value = mock_bedrock_client

    aip_model_id = "arn:aws:bedrock:us-west-2:123456789012:application-inference-profile/my-profile"  # noqa: E501
    chat = ChatBedrock(
        model_id=aip_model_id,
        region_name="us-west-2",
        bedrock_client=mock_bedrock_client,
    )  # type: ignore[call-arg]

    mock_bedrock_client.get_inference_profile.assert_called_with(
        inferenceProfileIdentifier=aip_model_id
    )

    assert chat.beta_use_converse_api is False


@mock.patch("langchain_aws.chat_models.bedrock.create_aws_client")
def test_beta_use_converse_api_with_inference_profile_as_nova_model(
    mock_create_aws_client,
):
    mock_bedrock_client = mock.MagicMock()
    mock_bedrock_client.get_inference_profile.return_value = {
        "models": [
            {
                "modelArn": "arn:aws:bedrock:us-west-2::foundation-model/amazon.nova-micro-v1:0"  # noqa: E501
            }
        ]
    }
    mock_create_aws_client.return_value = mock_bedrock_client

    aip_model_id = "arn:aws:bedrock:us-west-2:123456789012:application-inference-profile/my-profile"  # noqa: E501
    chat = ChatBedrock(
        model_id=aip_model_id,
        region_name="us-west-2",
        bedrock_client=mock_bedrock_client,
    )  # type: ignore[call-arg]

    mock_bedrock_client.get_inference_profile.assert_called_with(
        inferenceProfileIdentifier=aip_model_id
    )

    assert chat.beta_use_converse_api is True


@pytest.mark.parametrize(
    "model_id, provider, expected_provider, expectation, region_name",
    [
        (
            "amer.amazon.nova-pro-v1:0",
            None,
            "amazon",
            nullcontext(),
            "us-west-2",
        ),
        (
            "jp.anthropic.claude-sonnet-4-5-20250929-v1:0",
            None,
            "anthropic",
            nullcontext(),
            "ap-northeast-1",
        ),
        (
            "au.anthropic.claude-sonnet-4-5-20250929-v1:0",
            None,
            "anthropic",
            nullcontext(),
            "ap-southeast-2",
        ),
        (
            "global.anthropic.claude-sonnet-4-20250514-v1:0",
            None,
            "anthropic",
            nullcontext(),
            "us-west-2",
        ),
        (
            "eu.anthropic.claude-haiku-4-5-20251001-v1:0",
            None,
            "anthropic",
            nullcontext(),
            "eu-west-1",
        ),
        (
            "apac.anthropic.claude-3-5-sonnet-20240620-v1:0",
            None,
            "anthropic",
            nullcontext(),
            "ap-northeast-1",
        ),
        (
            "us-gov.anthropic.claude-3-5-sonnet-20240620-v1:0",
            None,
            "anthropic",
            nullcontext(),
            "us-gov-west-1",
        ),
        (
            "us.anthropic.claude-sonnet-4-20250514-v1:0",
            None,
            "anthropic",
            nullcontext(),
            "us-west-2",
        ),
        ("meta.llama3-1-405b-instruct-v1:0", None, "meta", nullcontext(), "us-west-2"),
        (
            "arn:aws:bedrock:us-east-1::custom-model/cohere.command-r-v1:0/MyCustomModel2",
            "cohere",
            "cohere",
            nullcontext(),
            "us-east-1",
        ),
        (
            "arn:aws:bedrock:us-east-1::custom-model/cohere.command-r-v1:0/MyCustomModel2",
            None,
            "cohere",
            pytest.raises(ValueError),
            "us-east-1",
        ),
    ],
)
def test__get_provider(
    model_id, provider, expected_provider, expectation, region_name
) -> None:
    llm = ChatBedrock(model_id=model_id, provider=provider, region_name=region_name)
    with expectation:
        assert llm._get_provider() == expected_provider


@mock.patch.dict(os.environ, {"AWS_REGION": "us-west-1"})
def test_chat_bedrock_different_regions() -> None:
    region = "ap-south-2"
    llm = ChatBedrock(
        model_id="anthropic.claude-3-sonnet-20240229-v1:0", region_name=region
    )
    assert llm.region_name == region


def test__format_anthropic_messages_with_thinking_blocks() -> None:
    """Test that thinking blocks are correctly formatted and preserved in messages."""
    system = SystemMessage("System instruction")  # type: ignore[misc]
    human = HumanMessage("What is the weather in NYC?")  # type: ignore[misc]
    ai = AIMessage(  # type: ignore[misc]
        "",
        additional_kwargs={
            "thinking": {
                "text": "I need to check the weather in NYC.",
                "signature": "SIG123",
            }
        },
        tool_calls=[{"name": "get_weather", "id": "1", "args": {"city": "nyc"}}],
    )
    tool = ToolMessage(  # type: ignore[misc]
        "It might be cloudy in nyc",
        tool_call_id="1",
    )

    messages = [system, human, ai, tool]
    expected_system, expected_messages = (
        "System instruction",
        [
            {"role": "user", "content": "What is the weather in NYC?"},
            {
                "role": "assistant",
                "content": [
                    {
                        "type": "thinking",
                        "thinking": "I need to check the weather in NYC.",
                        "signature": "SIG123",
                    },
                    {
                        "type": "tool_use",
                        "name": "get_weather",
                        "id": "1",
                        "input": {"city": "nyc"},
                    },
                ],
            },
            {
                "role": "user",
                "content": [
                    {
                        "type": "tool_result",
                        "content": "It might be cloudy in nyc",
                        "tool_use_id": "1",
                    }
                ],
            },
        ],
    )

    actual_system, actual_messages = _format_anthropic_messages(messages)
    assert expected_system == actual_system
    assert expected_messages == actual_messages


def test__format_anthropic_messages_with_image_conversion_in_tool() -> None:
    """Test that ToolMessage with OpenAI-style image content is correctly
    converted to Anthropic format."""
    # Create a dummy base64 image string
    dummy_base64_image = (
        "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNk+M9QDwADhgGAWjR9aw"
        "AAAABJRU5ErkJggg=="
    )

    messages = [
        ToolMessage(  # type: ignore[misc]
            content=[
                {
                    "type": "image_url",
                    "image_url": {"url": f"data:image/png;base64,{dummy_base64_image}"},
                }
            ],
            tool_call_id="test_tool_call_123",
        ),
        HumanMessage("What do you see in the image?"),  # type: ignore[misc]
    ]

    expected = [
        {
            "role": "user",
            "content": [
                {
                    "type": "tool_result",
                    "tool_use_id": "test_tool_call_123",
                    "content": [
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": "image/png",
                                "data": dummy_base64_image,
                            },
                        }
                    ],
                },
                {"type": "text", "text": "What do you see in the image?"},
            ],
        }
    ]

    _, actual = _format_anthropic_messages(messages)
    assert expected == actual


def test__format_anthropic_messages_strips_extra_fields_from_tool_result_text() -> None:
    """Test that extra fields are stripped from text content block inside
    tool_result content."""
    messages = [
        ToolMessage(  # type: ignore[misc]
            content=[
                {
                    "type": "text",
                    "text": "tool output",
                    "id": "lc_12345",
                }
            ],
            tool_call_id="test_tool_call_123",
        ),
        HumanMessage("What happened?"),  # type: ignore[misc]
    ]

    _, actual = _format_anthropic_messages(messages)
    tool_result = actual[0]["content"][0]
    assert tool_result["type"] == "tool_result"
    assert tool_result["tool_use_id"] == "test_tool_call_123"

    # The text block inside the tool_result should only have "type" and "text"
    text_block = tool_result["content"][0]
    assert text_block == {"type": "text", "text": "tool output"}
    assert "id" not in text_block


def test__convert_messages_to_prompt_anthropic_message_is_none() -> None:
    messages = None
    assert convert_messages_to_prompt_anthropic(messages) == ""


def test__convert_messages_to_prompt_anthropic_message_is_empty() -> None:
    messages = []
    assert convert_messages_to_prompt_anthropic(messages) == ""


def test__format_anthropic_messages_with_thinking_in_content_blocks() -> None:
    """Test that thinking blocks in content are correctly ordered (first) in
    messages."""
    system = SystemMessage("System instruction")  # type: ignore[misc]
    human = HumanMessage("What is the weather in NYC?")  # type: ignore[misc]

    # Create AIMessage with content list that has thinking block not at the start
    ai = AIMessage(  # type: ignore[misc]
        [
            {"type": "text", "text": "Let me check the weather."},
            {
                "type": "thinking",
                "thinking": "I should use the get_weather tool.",
                "signature": "SIG456",
            },
            {
                "type": "tool_use",
                "id": "tool1",
                "name": "get_weather",
                "input": {"city": "nyc"},
            },
        ],
    )
    tool = ToolMessage("It might be cloudy in nyc", tool_call_id="tool1")  # type: ignore[misc]

    messages = [system, human, ai, tool]
    expected_system, expected_messages = (
        "System instruction",
        [
            {"role": "user", "content": "What is the weather in NYC?"},
            {
                "role": "assistant",
                "content": [
                    {
                        "type": "thinking",
                        "thinking": "I should use the get_weather tool.",
                        "signature": "SIG456",
                    },
                    {"type": "text", "text": "Let me check the weather."},
                    {
                        "type": "tool_use",
                        "id": "tool1",
                        "name": "get_weather",
                        "input": {"city": "nyc"},
                    },
                ],
            },
            {
                "role": "user",
                "content": [
                    {
                        "type": "tool_result",
                        "content": "It might be cloudy in nyc",
                        "tool_use_id": "tool1",
                    }
                ],
            },
        ],
    )

    actual_system, actual_messages = _format_anthropic_messages(messages)
    assert expected_system == actual_system

    # Verify thinking blocks are placed first in the content array
    assert actual_messages[1]["content"][0]["type"] == "thinking"
    assert expected_messages == actual_messages


def test__format_anthropic_messages_after_tool_use_no_thinking() -> None:
    """Test message formatting for assistant responses after tool use (which
    shouldn't have thinking)."""
    system = SystemMessage("System instruction")  # type: ignore[misc]
    human = HumanMessage("What is the weather in NYC?")  # type: ignore[misc]

    # First assistant turn with thinking and tool use
    assistant1 = AIMessage(  # type: ignore[misc]
        "",
        additional_kwargs={
            "thinking": {
                "text": "I need to check the weather in NYC.",
                "signature": "SIG123",
            }
        },
        tool_calls=[{"name": "get_weather", "id": "1", "args": {"city": "nyc"}}],
    )

    # Tool result from user
    tool_result = ToolMessage("It might be cloudy in nyc", tool_call_id="1")  # type: ignore[misc]

    # Final assistant response without thinking
    assistant2 = AIMessage("Based on the data, it's cloudy in NYC.")  # type: ignore[misc]

    messages = [system, human, assistant1, tool_result, assistant2]
    _, actual_messages = _format_anthropic_messages(messages)

    # Check that the final assistant message has no thinking blocks
    assert len(actual_messages) == 4  # system isn't included in the array
    assert actual_messages[3]["role"] == "assistant"

    # The content should be a list with a single text block
    assert isinstance(actual_messages[3]["content"], list)
    assert len(actual_messages[3]["content"]) == 1
    assert actual_messages[3]["content"][0]["type"] == "text"
    assert (
        actual_messages[3]["content"][0]["text"]
        == "Based on the data, it's cloudy in NYC."
    )

    # Verify no thinking blocks in the final message
    assert not any(
        block.get("type") in ["thinking", "redacted_thinking"]
        for block in actual_messages[3]["content"]
    )


def test__format_anthropic_messages_tool_result_ordering() -> None:
    """Test that tool type content blocks in UserMessage are moved to the beginning."""
    human_message = HumanMessage(  # type: ignore[misc]
        [
            {"type": "text", "text": "I have a question about the data."},
            {
                "type": "tool_result",
                "content": "Data analysis result",
                "tool_use_id": "tool1",
            },
            {"type": "text", "text": "Can you explain this result?"},
        ]
    )

    system = SystemMessage("You are a helpful assistant")  # type: ignore[misc]
    ai = AIMessage("Let me think about that")  # type: ignore[misc]

    messages = [system, human_message, ai]

    _, formatted_messages = _format_anthropic_messages(messages)

    user_content = formatted_messages[0]["content"]

    assert user_content[0]["type"] == "tool_result"
    assert user_content[0]["content"] == "Data analysis result"
    assert user_content[1]["type"] == "text"
    assert user_content[1]["text"] == "I have a question about the data."
    assert user_content[2]["type"] == "text"
    assert user_content[2]["text"] == "Can you explain this result?"


def test__format_anthropic_messages_tool_use_ordering() -> None:
    """Test that tool type content blocks in AssistantMessage are always moved to
    the end."""
    ai_message = AIMessage(  # type: ignore[misc]
        [
            {"type": "text", "text": "Let me analyze this for you."},
            {
                "type": "tool_use",
                "name": "data_analyzer",
                "id": "tool1",
                "input": {"data": "sample_data"},
            },
            {"type": "text", "text": "This will help us understand the pattern."},
        ]
    )

    system = SystemMessage("You are a helpful assistant")  # type: ignore[misc]
    human = HumanMessage("Can you analyze this data?")  # type: ignore[misc]

    messages = [system, human, ai_message]

    _, formatted_messages = _format_anthropic_messages(messages)

    assistant_content = formatted_messages[1]["content"]

    assert assistant_content[0]["type"] == "text"
    assert assistant_content[0]["text"] == "Let me analyze this for you."
    assert assistant_content[1]["type"] == "text"
    assert assistant_content[1]["text"] == "This will help us understand the pattern."
    assert assistant_content[2]["type"] == "tool_use"
    assert assistant_content[2]["name"] == "data_analyzer"


def test__format_anthropic_messages_preserves_content_order() -> None:
    """Test that _format_anthropic_messages preserves the original order of mixed
    text and image content."""
    content = [
        {"type": "text", "text": "Some text..."},
        {
            "type": "image_url",
            "image_url": {"url": "data:image/jpeg;base64,1337C0DE"},
        },
        {"type": "text", "text": "Caption for 1st image..."},
        {
            "type": "image_url",
            "image_url": {"url": "data:image/jpeg;base64,FACADE42"},
        },
        {"type": "text", "text": "Another caption for 2nd image..."},
        {"type": "text", "text": "Now, analyze the following image..."},
        {
            "type": "image_url",
            "image_url": {"url": "data:image/jpeg;base64,C0FFEE42"},
        },
    ]

    messages = [HumanMessage(content=content)]

    _, formatted_messages = _format_anthropic_messages(messages)

    assert len(formatted_messages) == 1
    assert formatted_messages[0]["role"] == "user"

    user_content = formatted_messages[0]["content"]

    # The expected order should preserve the original interleaved sequence:
    # text -> image -> text -> image -> text -> text -> image
    expected_sequence = [
        ("text", "Some text..."),
        ("image", "1337C0DE"),
        ("text", "Caption for 1st image..."),
        ("image", "FACADE42"),
        ("text", "Another caption for 2nd image..."),
        ("text", "Now, analyze the following image..."),
        ("image", "C0FFEE42"),
    ]

    actual_sequence = []
    for item in user_content:
        if item["type"] == "text":
            actual_sequence.append(("text", item["text"]))
        elif item["type"] == "image":
            actual_sequence.append(("image", item["source"]["data"]))

    assert actual_sequence == expected_sequence, (
        f"Content order was not preserved. Expected: {expected_sequence}, "
        f"but got: {actual_sequence}"
    )


@pytest.mark.parametrize(
    "model_id, base_model_id, provider, expected_format_marker",
    [
        (
            "arn:aws:bedrock:us-west-2::custom-model/meta.llama3-8b-instruct-v1:0/MyModel",
            "meta.llama3-8b-instruct-v1:0",
            "meta",
            "<|begin_of_text|>",
        ),
        (
            "arn:aws:bedrock:us-west-2::custom-model/meta.llama2-70b-chat-v1/MyModel",
            "meta.llama2-70b-chat-v1",
            "meta",
            "[INST]",
        ),
        (
            "meta.llama2-70b-chat-v1",
            "meta.llama3-8b-instruct-v1:0",
            "meta",
            "<|begin_of_text|>",
        ),
        (
            "arn:aws:sagemaker:us-west-2::endpoint/endpoint-quick-start-xxxxx",
            "deepseek.r1-v1:0",
            "deepseek",
            "<|begin_of_sentence|>",
        ),
    ],
)
def test_chat_prompt_adapter_with_model_detection(
    model_id, base_model_id, provider, expected_format_marker
):
    """Test that ChatPromptAdapter correctly formats prompts when base_model is provided."""  # noqa: E501
    messages = [
        SystemMessage(content="You are a helpful assistant"),
        HumanMessage(content="Hello"),
    ]

    chat = ChatBedrock(
        model_id=model_id,
        base_model_id=base_model_id,
        provider=provider,
        region_name="us-west-2",
    )

    model_name = chat._get_base_model()
    provider_name = chat._get_provider()

    prompt = ChatPromptAdapter.convert_messages_to_prompt(
        provider=provider_name, messages=messages, model=model_name
    )

    assert expected_format_marker in prompt


def test__format_anthropic_messages_empty_content_fix() -> None:
    """
    Test that empty content arrays are handled correctly to prevent ValidationException.
    """
    messages = [
        HumanMessage("What is the capital of India?"),  # type: ignore[misc]
        AIMessage([{"type": "text", "text": ""}]),  # type: ignore[misc]
    ]

    system, formatted_messages = _format_anthropic_messages(messages)

    assert len(formatted_messages) == 2
    ai_content = formatted_messages[1]["content"]
    assert isinstance(ai_content, list)
    assert len(ai_content) > 0  # Should not be empty
    assert ai_content[0]["type"] == "text"
    assert ai_content[0]["text"] == "."


def test__format_anthropic_messages_whitespace_only_content() -> None:
    """Test that whitespace-only content is handled correctly."""
    messages = [
        HumanMessage("What is the capital of India?"),  # type: ignore[misc]
        AIMessage([{"type": "text", "text": "   \n  \t  "}]),  # type: ignore[misc]
    ]

    system, formatted_messages = _format_anthropic_messages(messages)

    assert len(formatted_messages) == 2
    ai_content = formatted_messages[1]["content"]
    assert isinstance(ai_content, list)
    assert len(ai_content) > 0
    assert ai_content[0]["type"] == "text"
    assert ai_content[0]["text"] == "."


def test__format_anthropic_messages_empty_string_content() -> None:
    """Test that empty string content is handled correctly."""
    messages = [
        HumanMessage("What is the capital of India?"),  # type: ignore[misc]
        AIMessage(""),  # type: ignore[misc]
    ]

    system, formatted_messages = _format_anthropic_messages(messages)

    assert len(formatted_messages) == 2
    ai_content = formatted_messages[1]["content"]
    assert isinstance(ai_content, list)
    assert len(ai_content) > 0
    assert ai_content[0]["type"] == "text"
    assert ai_content[0]["text"] == "."


def test__format_anthropic_messages_mixed_empty_content() -> None:
    """Test that mixed content with some empty blocks is handled correctly."""
    messages = [
        HumanMessage("What is the capital of India?"),  # type: ignore[misc]
        AIMessage(
            [  # type: ignore[misc]
                {"type": "text", "text": ""},
                {"type": "text", "text": "   "},
                {"type": "text", "text": ""},
            ]
        ),
    ]

    system, formatted_messages = _format_anthropic_messages(messages)

    # Verify that the content is not empty even when all text blocks are filtered out
    assert len(formatted_messages) == 2
    ai_content = formatted_messages[1]["content"]
    assert isinstance(ai_content, list)
    assert len(ai_content) > 0
    assert ai_content[0]["type"] == "text"
    assert ai_content[0]["text"] == "."


def test__format_anthropic_messages_mixed_type_blocks_and_empty_content() -> None:
    """Test that empty blocks mixed with non-text type blocks is handled correctly."""
    messages = [
        AIMessage(
            [  # type: ignore[misc]
                {"type": "text", "text": "\n\t"},
                {
                    "type": "tool_use",
                    "id": "tool_call1",
                    "input": {"arg1": "val1"},
                    "name": "tool1",
                },
            ]
        )
    ]

    expected_content = [
        {
            "role": "assistant",
            "content": [
                {
                    "type": "tool_use",
                    "id": "tool_call1",
                    "input": {"arg1": "val1"},
                    "name": "tool1",
                }
            ],
        }
    ]

    system, formatted_messages = _format_anthropic_messages(messages)

    assert len(formatted_messages) == 1
    assert formatted_messages == expected_content


def test_model_kwargs() -> None:
    """Test we can transfer unknown params to model_kwargs."""
    llm = ChatBedrock(
        model_id="my-model",
        region_name="us-west-2",
        model_kwargs={"foo": "bar"},
    )
    assert llm.model_id == "my-model"
    assert llm.region_name == "us-west-2"
    assert llm.model_kwargs == {"foo": "bar"}

    with pytest.warns(match="transferred to model_kwargs"):
        llm = ChatBedrock(
            model_id="my-model",
            region_name="us-west-2",
            foo="bar",
        )
    assert llm.model_id == "my-model"
    assert llm.region_name == "us-west-2"
    assert llm.model_kwargs == {"foo": "bar"}

    with pytest.warns(match="transferred to model_kwargs"):
        llm = ChatBedrock(
            model_id="my-model",
            region_name="us-west-2",
            foo="bar",
            model_kwargs={"baz": "qux"},
        )
    assert llm.model_id == "my-model"
    assert llm.region_name == "us-west-2"
    assert llm.model_kwargs == {"foo": "bar", "baz": "qux"}

    # For backward compatibility, test that we don't transfer known parameters out
    # of model_kwargs
    llm = ChatBedrock(
        model_id="my-model",
        region_name="us-west-2",
        model_kwargs={"stop_sequences": ["test"]},
    )
    assert llm.model_kwargs == {"stop_sequences": ["test"]}
    assert llm.stop_sequences is None


def test__format_anthropic_messages_strips_trailing_whitespace_string() -> None:
    """Test that _format_anthropic_messages strips trailing whitespace from AIMessage string content."""  # noqa: E501
    messages = [
        SystemMessage(content="System message"),
        HumanMessage(content="Human message"),
        AIMessage(content="AI message with trailing whitespace    \n  \t  "),
    ]

    _, formatted_messages = _format_anthropic_messages(messages)

    assert (
        formatted_messages[1]["content"][0]["text"]
        == "AI message with trailing whitespace"
    )


def test__format_anthropic_messages_strips_trailing_whitespace_blocks() -> None:
    """Test that _format_anthropic_messages strips trailing whitespace from AIMessage dict content."""  # noqa: E501
    messages = [
        SystemMessage(content="System message"),
        HumanMessage(content="Human message"),
        AIMessage(
            content=[
                {
                    "type": "text",
                    "text": "AI message with trailing whitespace    \n  \t  ",
                },
                {"type": "text", "text": "Another text block with whitespace  \n "},
            ]
        ),
    ]

    _, formatted_messages = _format_anthropic_messages(messages)

    assert (
        formatted_messages[1]["content"][0]["text"]
        == "AI message with trailing whitespace"
    )
    assert (
        formatted_messages[1]["content"][1]["text"]
        == "Another text block with whitespace"
    )


def test__format_anthropic_messages_preserves_whitespace_non_last_aimessage_string() -> (  # noqa: E501
    None
):
    """Test that _format_anthropic_messages preserves trailing whitespace in non-last AIMessages."""  # noqa: E501
    messages = [
        SystemMessage(content="System message"),
        HumanMessage(content="First human message"),
        AIMessage(content="AI message with trailing whitespace    \n  \t  "),
        HumanMessage(content="Second human message"),
        AIMessage(content="Final AI message"),
    ]

    _, formatted_messages = _format_anthropic_messages(messages)

    assert (
        formatted_messages[1]["content"][0]["text"]
        == "AI message with trailing whitespace    \n  \t  "
    )


def test__format_anthropic_messages_preserves_whitespace_non_last_aimessage_blocks() -> (  # noqa: E501
    None
):
    """Test that _format_anthropic_messages preserves trailing whitespace in non-last AIMessages."""  # noqa: E501
    messages = [
        SystemMessage(content="System message"),
        HumanMessage(content="First human message"),
        AIMessage(
            content=[
                {
                    "type": "text",
                    "text": "AI message with trailing whitespace    \n  \t  ",
                },
            ]
        ),
        HumanMessage(content="Second human message"),
    ]

    _, formatted_messages = _format_anthropic_messages(messages)

    assert (
        formatted_messages[1]["content"][0]["text"]
        == "AI message with trailing whitespace    \n  \t  "
    )


@patch("langchain_aws.llms.bedrock.create_aws_client")
def test_bedrock_client_inherits_from_runtime_client(
    mock_create_client: MagicMock,
) -> None:
    """Test that bedrock_client inherits region and config from runtime client."""
    mock_runtime_client = MagicMock()
    mock_bedrock_client = MagicMock()

    mock_runtime_client.meta.region_name = "us-west-2"
    mock_client_config = MagicMock()
    mock_runtime_client._client_config = mock_client_config

    def side_effect(service_name: str, **kwargs: Any) -> MagicMock:
        if service_name == "bedrock":
            return mock_bedrock_client
        elif service_name == "bedrock-runtime":
            return mock_runtime_client
        return MagicMock()

    mock_create_client.side_effect = side_effect

    with mock.patch.dict(os.environ, {}, clear=False):
        os.environ.pop("AWS_BEARER_TOKEN_BEDROCK", None)
        ChatBedrock(
            model="us.meta.llama3-3-70b-instruct-v1:0", client=mock_runtime_client
        )

        mock_create_client.assert_called_with(
            region_name="us-west-2",
            credentials_profile_name=None,
            aws_access_key_id=None,
            aws_secret_access_key=None,
            aws_session_token=None,
            endpoint_url=None,
            config=mock_client_config,
            service_name="bedrock",
            api_key=None,
        )


@patch("langchain_aws.llms.bedrock.create_aws_client")
def test_bedrock_client_uses_explicit_values_over_runtime_client(
    mock_create_client: MagicMock,
) -> None:
    """Test that explicitly provided values override those from runtime client."""
    mock_runtime_client = MagicMock()
    mock_bedrock_client = MagicMock()

    mock_runtime_client.meta.region_name = "us-west-2"
    mock_runtime_config = MagicMock()
    mock_runtime_client._client_config = mock_runtime_config

    explicit_config = MagicMock()

    def side_effect(service_name: str, **kwargs: Any) -> MagicMock:
        if service_name == "bedrock":
            return mock_bedrock_client
        elif service_name == "bedrock-runtime":
            return mock_runtime_client
        return MagicMock()

    mock_create_client.side_effect = side_effect

    with mock.patch.dict(os.environ, {}, clear=False):
        os.environ.pop("AWS_BEARER_TOKEN_BEDROCK", None)
        ChatBedrock(
            model="us.meta.llama3-3-70b-instruct-v1:0",
            client=mock_runtime_client,
            region="us-east-1",
            config=explicit_config,
        )

        mock_create_client.assert_called_with(
            region_name="us-east-1",
            credentials_profile_name=None,
            aws_access_key_id=None,
            aws_secret_access_key=None,
            aws_session_token=None,
            endpoint_url=None,
            config=explicit_config,
            service_name="bedrock",
            api_key=None,
        )


def test_get_num_tokens_from_messages_with_base_messages():
    """Test get_num_tokens_from_messages with BaseMessage objects."""
    mock_client = MagicMock()
    mock_client.count_tokens.return_value = {"inputTokens": 20}

    chat = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        region="us-west-2",
        max_tokens=4096,
    )
    chat.client = mock_client
    messages = [
        SystemMessage(content="You are a helpful assistant."),
        HumanMessage(content="Why did the chicken cross the road?"),
    ]

    token_count = chat.get_num_tokens_from_messages(messages)
    assert token_count == 20

    mock_client.count_tokens.assert_called_once()
    call_args = mock_client.count_tokens.call_args[1]
    assert call_args["modelId"] == "anthropic.claude-sonnet-4-5-20250929-v1:0"

    actual_input_body = json.loads(call_args["input"]["invokeModel"]["body"])
    expected_input_body = {
        "anthropic_version": "bedrock-2023-05-31",
        "max_tokens": 4096,
        "system": "You are a helpful assistant.",
        "messages": [
            {"role": "user", "content": "Why did the chicken cross the road?"}
        ],
    }

    assert actual_input_body == expected_input_body


@patch("langchain_aws.chat_models.bedrock.create_aws_client")
@patch("langchain_aws.chat_models.bedrock.count_tokens_api_supported_for_model")
def test_get_num_tokens_from_messages_fallback(
    mock_count_tokens_api_supported, mock_create_aws_client
):
    """Test get_num_tokens_from_messages falls back to parent if API not supported."""
    mock_count_tokens_api_supported.return_value = False
    mock_client = MagicMock()
    mock_create_aws_client.return_value = mock_client

    chat = ChatBedrock(
        model="us.anthropic.claude-3-sonnet-20240229-v1:0",
        region="us-west-2",
    )

    with patch.object(
        BaseChatModel, "get_num_tokens_from_messages", return_value=24
    ) as mock_parent:
        messages = [
            SystemMessage(content="You are a helpful assistant."),
            HumanMessage(content="Why did the chicken cross the road?"),
        ]

        token_count = chat.get_num_tokens_from_messages(messages)
        assert token_count == 24
        mock_parent.assert_called_once()
        mock_client.count_tokens.assert_not_called()


def test_service_tier_parameter() -> None:
    """Test that service_tier parameter is correctly set."""
    llm = ChatBedrock(
        model_id="anthropic.claude-3-sonnet-20240229-v1:0",
        region_name="us-west-2",
        service_tier="priority",
    )
    assert llm.service_tier == "priority"


@pytest.mark.parametrize(
    "service_tier",
    ["priority", "default", "flex", "reserved"],
)
def test_service_tier_valid_values(service_tier: str) -> None:
    """Test that all valid service_tier values are accepted."""
    llm = ChatBedrock(
        model_id="anthropic.claude-3-sonnet-20240229-v1:0",
        region_name="us-west-2",
        service_tier=service_tier,  # type: ignore[arg-type]
    )
    assert llm.service_tier == service_tier


def test_service_tier_passed_to_invoke_model() -> None:
    """Test that service_tier is passed to the invoke_model API call."""
    mocked_client = MagicMock()

    # Create mock response
    mock_response = MagicMock()
    mock_response.get.return_value.read.return_value = json.dumps(
        {"completion": "Hello!", "stop_reason": "end_turn"}
    ).encode()
    mock_response.get.side_effect = lambda key, default=None: {
        "body": mock_response.get.return_value,
        "ResponseMetadata": {
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "10",
                "x-amzn-bedrock-output-token-count": "5",
            }
        },
    }.get(key, default)

    mocked_client.invoke_model.return_value = mock_response

    llm = ChatBedrock(
        client=mocked_client,
        model_id="anthropic.claude-v2",
        region_name="us-west-2",
        service_tier="priority",
    )

    messages = [HumanMessage(content="Hi")]
    llm.invoke(messages)

    # Verify invoke_model was called with serviceTier in the correct format
    call_kwargs = mocked_client.invoke_model.call_args[1]
    assert call_kwargs["serviceTier"] == "priority"


def test_service_tier_passed_to_as_converse() -> None:
    """Test that service_tier is passed to _as_converse."""
    mocked_client = MagicMock()

    llm = ChatBedrock(
        client=mocked_client,
        model_id="amazon.nova-lite-v1:0",
        region_name="us-west-2",
        service_tier="flex",
    )

    converse_llm = llm._as_converse
    assert converse_llm.service_tier == "flex"


def test_system_prompt_cache_control_preserved() -> None:
    """Test that system prompt with cache_control is preserved as list."""
    mock_client = MagicMock()
    mock_response = MagicMock()
    mock_response.get.return_value.read.return_value = json.dumps(
        {
            "content": [{"type": "text", "text": "Hello!"}],
            "usage": {"input_tokens": 10, "output_tokens": 5},
            "stop_reason": "end_turn",
        }
    ).encode()
    mock_response.get.side_effect = lambda key, default=None: {
        "body": mock_response.get.return_value,
        "ResponseMetadata": {"HTTPHeaders": {}},
    }.get(key, default)
    mock_client.invoke_model.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    system = SystemMessage(
        [
            {
                "type": "text",
                "text": "You are helpful.",
                "cache_control": {"type": "ephemeral"},
            }
        ]
    )
    llm.invoke([system, HumanMessage(content="Hi")])

    body = json.loads(mock_client.invoke_model.call_args[1]["body"])
    assert isinstance(body["system"], list)
    assert body["system"][0].get("cache_control") == {"type": "ephemeral"}


def test_system_prompt_list_format() -> None:
    """Test that list system prompts are passed through without conversion."""
    mock_client = MagicMock()
    mock_response = MagicMock()
    mock_response.get.return_value.read.return_value = json.dumps(
        {
            "content": [{"type": "text", "text": "Hello!"}],
            "usage": {"input_tokens": 10, "output_tokens": 5},
            "stop_reason": "end_turn",
        }
    ).encode()
    mock_response.get.side_effect = lambda key, default=None: {
        "body": mock_response.get.return_value,
        "ResponseMetadata": {"HTTPHeaders": {}},
    }.get(key, default)
    mock_client.invoke_model.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    system = SystemMessage(
        [
            {"type": "text", "text": "Block 1."},
            {"type": "text", "text": "Block 2."},
        ]
    )
    llm.invoke([system, HumanMessage(content="Hi")])

    body = json.loads(mock_client.invoke_model.call_args[1]["body"])
    assert isinstance(body["system"], list)
    assert len(body["system"]) == 2


def test_system_prompt_with_tools_prepends_block() -> None:
    """Test that tools are prepended as content block when system is a list."""
    mock_client = MagicMock()
    mock_response = MagicMock()
    mock_response.get.return_value.read.return_value = json.dumps(
        {
            "content": [{"type": "text", "text": "Hello!"}],
            "usage": {"input_tokens": 10, "output_tokens": 5},
            "stop_reason": "end_turn",
        }
    ).encode()
    mock_response.get.side_effect = lambda key, default=None: {
        "body": mock_response.get.return_value,
        "ResponseMetadata": {"HTTPHeaders": {}},
    }.get(key, default)
    mock_client.invoke_model.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    llm.system_prompt_with_tools = "You have tools."
    system = SystemMessage(
        [
            {
                "type": "text",
                "text": "Be helpful.",
                "cache_control": {"type": "ephemeral"},
            }
        ]
    )
    llm.invoke([system, HumanMessage(content="Hi")])

    body = json.loads(mock_client.invoke_model.call_args[1]["body"])
    assert isinstance(body["system"], list)
    assert len(body["system"]) == 2
    assert body["system"][0]["text"] == "You have tools."
    assert "cache_control" not in body["system"][0]
    assert body["system"][1].get("cache_control") == {"type": "ephemeral"}


def test_system_prompt_string_format() -> None:
    """Test that string system prompts still work."""
    mock_client = MagicMock()
    mock_response = MagicMock()
    mock_response.get.return_value.read.return_value = json.dumps(
        {
            "content": [{"type": "text", "text": "Hello!"}],
            "usage": {"input_tokens": 10, "output_tokens": 5},
            "stop_reason": "end_turn",
        }
    ).encode()
    mock_response.get.side_effect = lambda key, default=None: {
        "body": mock_response.get.return_value,
        "ResponseMetadata": {"HTTPHeaders": {}},
    }.get(key, default)
    mock_client.invoke_model.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    llm.invoke([SystemMessage("You are helpful."), HumanMessage(content="Hi")])  # type: ignore[misc]

    body = json.loads(mock_client.invoke_model.call_args[1]["body"])
    assert isinstance(body["system"], str)


def test_stream_system_prompt_cache_control() -> None:
    """Test that cache_control is preserved in streaming."""
    mock_client = MagicMock()

    def stream_gen():
        yield {
            "chunk": {
                "bytes": json.dumps(
                    {
                        "type": "content_block_delta",
                        "delta": {"type": "text_delta", "text": "Hi"},
                    }
                ).encode()
            }
        }
        yield {
            "chunk": {
                "bytes": json.dumps(
                    {
                        "type": "message_stop",
                        "amazon-bedrock-invocationMetrics": {
                            "inputTokenCount": 10,
                            "outputTokenCount": 5,
                        },
                    }
                ).encode()
            }
        }

    mock_response = MagicMock()
    mock_response.get.return_value = stream_gen()
    mock_client.invoke_model_with_response_stream.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    system = SystemMessage(
        [
            {
                "type": "text",
                "text": "You are helpful.",
                "cache_control": {"type": "ephemeral"},
            }
        ]
    )
    list(llm.stream([system, HumanMessage(content="Hi")]))

    body = json.loads(
        mock_client.invoke_model_with_response_stream.call_args[1]["body"]
    )
    assert isinstance(body["system"], list)
    assert body["system"][0].get("cache_control") == {"type": "ephemeral"}


@pytest.mark.asyncio
async def test_astream_system_prompt_cache_control() -> None:
    """Test that cache_control is preserved in async streaming."""
    mock_client = MagicMock()

    def stream_gen():
        yield {
            "chunk": {
                "bytes": json.dumps(
                    {
                        "type": "content_block_delta",
                        "delta": {"type": "text_delta", "text": "Hi"},
                    }
                ).encode()
            }
        }
        yield {
            "chunk": {
                "bytes": json.dumps(
                    {
                        "type": "message_stop",
                        "amazon-bedrock-invocationMetrics": {
                            "inputTokenCount": 10,
                            "outputTokenCount": 5,
                        },
                    }
                ).encode()
            }
        }

    mock_response = MagicMock()
    mock_response.get.return_value = stream_gen()
    mock_client.invoke_model_with_response_stream.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    system = SystemMessage(
        [
            {
                "type": "text",
                "text": "You are helpful.",
                "cache_control": {"type": "ephemeral"},
            }
        ]
    )
    async for _ in llm.astream([system, HumanMessage(content="Hi")]):
        pass

    body = json.loads(
        mock_client.invoke_model_with_response_stream.call_args[1]["body"]
    )
    assert isinstance(body["system"], list)
    assert body["system"][0].get("cache_control") == {"type": "ephemeral"}


def test_cache_control_kwarg_applied_to_string_content() -> None:
    """Test that cache_control kwarg converts string content to block."""
    mock_client = MagicMock()
    mock_response = MagicMock()
    mock_response.get.return_value.read.return_value = json.dumps(
        {
            "content": [{"type": "text", "text": "Hello!"}],
            "usage": {"input_tokens": 10, "output_tokens": 5},
            "stop_reason": "end_turn",
        }
    ).encode()
    mock_response.get.side_effect = lambda key, default=None: {
        "body": mock_response.get.return_value,
        "ResponseMetadata": {"HTTPHeaders": {}},
    }.get(key, default)
    mock_client.invoke_model.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    llm.invoke(
        [SystemMessage("You are helpful."), HumanMessage(content="Hi")],
        cache_control={"type": "ephemeral", "ttl": "5m"},
    )

    body = json.loads(mock_client.invoke_model.call_args[1]["body"])
    last_msg = body["messages"][-1]
    # String content should be converted to list with cache_control
    assert isinstance(last_msg["content"], list)
    assert last_msg["content"][0] == {
        "type": "text",
        "text": "Hi",
        "cache_control": {"type": "ephemeral", "ttl": "5m"},
    }


def test_cache_control_kwarg_applied_to_list_content() -> None:
    """Test that cache_control kwarg is applied to last block."""
    mock_client = MagicMock()
    mock_response = MagicMock()
    mock_response.get.return_value.read.return_value = json.dumps(
        {
            "content": [{"type": "text", "text": "Hello!"}],
            "usage": {"input_tokens": 10, "output_tokens": 5},
            "stop_reason": "end_turn",
        }
    ).encode()
    mock_response.get.side_effect = lambda key, default=None: {
        "body": mock_response.get.return_value,
        "ResponseMetadata": {"HTTPHeaders": {}},
    }.get(key, default)
    mock_client.invoke_model.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    llm.invoke(
        [
            SystemMessage("You are helpful."),
            HumanMessage(content="Hello"),
            AIMessage(content="Hi there!"),
            HumanMessage(content="How are you?"),
        ],
        cache_control={"type": "ephemeral"},
    )

    body = json.loads(mock_client.invoke_model.call_args[1]["body"])
    last_msg = body["messages"][-1]
    content = last_msg["content"]
    if isinstance(content, list):
        last_block = next(b for b in reversed(content) if isinstance(b, dict))
        assert last_block["cache_control"] == {"type": "ephemeral"}
    else:
        pytest.fail("Expected list content after cache_control applied")


def test_cache_control_kwarg_not_in_api_body() -> None:
    """Test that cache_control is popped from params, not sent to API."""
    mock_client = MagicMock()
    mock_response = MagicMock()
    mock_response.get.return_value.read.return_value = json.dumps(
        {
            "content": [{"type": "text", "text": "Hello!"}],
            "usage": {"input_tokens": 10, "output_tokens": 5},
            "stop_reason": "end_turn",
        }
    ).encode()
    mock_response.get.side_effect = lambda key, default=None: {
        "body": mock_response.get.return_value,
        "ResponseMetadata": {"HTTPHeaders": {}},
    }.get(key, default)
    mock_client.invoke_model.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    llm.invoke(
        [SystemMessage("You are helpful."), HumanMessage(content="Hi")],
        cache_control={"type": "ephemeral"},
    )

    body = json.loads(mock_client.invoke_model.call_args[1]["body"])
    # cache_control should NOT leak as a top-level body parameter
    assert "cache_control" not in body


def test_stream_cache_control_kwarg_applied() -> None:
    """Test that cache_control kwarg works in streaming mode."""
    mock_client = MagicMock()

    def stream_gen():
        yield {
            "chunk": {
                "bytes": json.dumps(
                    {
                        "type": "content_block_delta",
                        "delta": {"type": "text_delta", "text": "Hi"},
                    }
                ).encode()
            }
        }
        yield {
            "chunk": {
                "bytes": json.dumps(
                    {
                        "type": "message_stop",
                        "amazon-bedrock-invocationMetrics": {
                            "inputTokenCount": 10,
                            "outputTokenCount": 5,
                        },
                    }
                ).encode()
            }
        }

    mock_response = MagicMock()
    mock_response.get.return_value = stream_gen()
    mock_client.invoke_model_with_response_stream.return_value = mock_response

    llm = ChatBedrock(
        client=mock_client,
        model_id="anthropic.claude-haiku-4-5-20251001-v1:0",
        region_name="us-west-2",
    )
    list(
        llm.stream(
            [SystemMessage("You are helpful."), HumanMessage(content="Hi")],
            cache_control={"type": "ephemeral", "ttl": "5m"},
        )
    )

    body = json.loads(
        mock_client.invoke_model_with_response_stream.call_args[1]["body"]
    )
    last_msg = body["messages"][-1]
    assert isinstance(last_msg["content"], list)
    assert last_msg["content"][0]["cache_control"] == {
        "type": "ephemeral",
        "ttl": "5m",
    }



import asyncio
import base64
import json
import logging
import os
import re
from typing import Any, Dict, Generator, List, Literal, Optional, Union

import numpy as np
from langchain_core.embeddings import Embeddings
from langchain_core.runnables.config import run_in_executor
from langchain_core.utils import secret_from_env
from pydantic import BaseModel, ConfigDict, Field, SecretStr, model_validator
from typing_extensions import Self

from langchain_aws.utils import create_aws_client

logger = logging.getLogger(__name__)

# Type alias for media input (base64 string, file path, S3 URI, or raw bytes)
MediaInput = Union[str, bytes]


class BedrockEmbeddings(BaseModel, Embeddings):
    """Bedrock embedding models.

    To authenticate, the AWS client uses the following methods to
    automatically load credentials:
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html

    If a specific credential profile should be used, you must pass
    the name of the profile from the ~/.aws/credentials file that is to be used.

    Make sure the credentials / roles used have the required policies to
    access the Bedrock service.

    Example:
        ```python
        from langchain_aws import BedrockEmbeddings

        embeddings = BedrockEmbeddings(
            credentials_profile_name="default",
            region_name="us-east-1",
            model_id="amazon.nova-2-multimodal-embeddings-v1:0",
            dimensions=256,
        )
        ```
    """

    client: Any = Field(default=None, exclude=True)
    """Bedrock client."""

    region_name: Optional[str] = None
    """The aws region e.g., `us-west-2`.

    Falls back to `AWS_REGION`/`AWS_DEFAULT_REGION` env variable or region
    specified in `~/.aws/config` in case it is not provided here.
    """

    credentials_profile_name: Optional[str] = None
    """The name of the profile in the `~/.aws/credentials` or `~/.aws/config` files,
    which has either access keys or role information specified.
    If not specified, the default credential profile or, if on an EC2 instance,
    credentials from IMDS will be used.

    See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html
    """

    aws_access_key_id: Optional[SecretStr] = Field(
        default_factory=secret_from_env("AWS_ACCESS_KEY_ID", default=None)
    )
    """AWS access key id.

    If provided, aws_secret_access_key must also be provided.
    If not specified, the default credential profile or, if on an EC2 instance,
    credentials from IMDS will be used.

    See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html

    If not provided, will be read from `AWS_ACCESS_KEY_ID` environment variable.
    """

    aws_secret_access_key: Optional[SecretStr] = Field(
        default_factory=secret_from_env("AWS_SECRET_ACCESS_KEY", default=None)
    )
    """AWS secret_access_key.

    If provided, aws_access_key_id must also be provided.
    If not specified, the default credential profile or, if on an EC2 instance,
    credentials from IMDS will be used.

    See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html

    If not provided, will be read from `AWS_SECRET_ACCESS_KEY` environment variable.

    """

    aws_session_token: Optional[SecretStr] = Field(
        default_factory=secret_from_env("AWS_SESSION_TOKEN", default=None)
    )
    """AWS session token.

    If provided, `aws_access_key_id` and `aws_secret_access_key` must also be
    provided.

    Not required unless using temporary credentials.

    See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html

    If not provided, will be read from `AWS_SESSION_TOKEN` environment variable.

    """

    model_id: str = "amazon.titan-embed-text-v1"
    """Id of the model to call, e.g., `'amazon.titan-embed-text-v1'`, this is
    equivalent to the `modelId` property in the list-foundation-models api

    """

    model_kwargs: Optional[Dict] = None
    """Keyword arguments to pass to the model."""

    provider: Optional[str] = None
    """Name of the provider, e.g., amazon, cohere, etc..
    If not specified, the provider will be inferred from the ``model_id``.

    """

    endpoint_url: Optional[str] = None
    """Needed if you don't want to default to `'us-east-1'` endpoint"""

    normalize: bool = False
    """Whether the embeddings should be normalized to unit vectors"""

    dimensions: Optional[int] = None
    """The number of dimensions for the output embeddings.

    Only supported by certain models (Titan v2, Cohere, Nova).
    If not specified, uses the model's default.
    """

    config: Any = None
    """An optional `botocore.config.Config` instance to pass to the client."""

    model_config = ConfigDict(
        extra="forbid",
        protected_namespaces=(),
    )

    @property
    def _inferred_provider(self) -> str:
        """Inferred provider of the model."""
        if self.provider:
            return self.provider

        regions = ("eu", "us", "us-gov", "apac", "sa", "amer", "global", "jp", "au")
        parts = self.model_id.split(".")
        return parts[1] if parts[0] in regions else parts[0]

    @property
    def _is_cohere_v4(self) -> bool:
        """Check if the model is Cohere Embed v4."""
        return "cohere.embed-v4" in self.model_id

    @property
    def _is_nova_embed(self) -> bool:
        """Check if the model is Amazon Nova Embed."""
        return (
            self._inferred_provider == "amazon"
            and "nova" in self.model_id
            and "embed" in self.model_id
        )

    @property
    def _is_titan_multimodal(self) -> bool:
        """Check if the model is Titan Embed Image (multimodal)."""
        return "titan-embed-image" in self.model_id

    @property
    def _is_marengo(self) -> bool:
        """Check if the model is TwelveLabs Marengo."""
        return "marengo" in self.model_id.lower()

    def _supports_image(self) -> bool:
        """Check if the model supports image embeddings."""
        return (
            self._is_titan_multimodal
            or self._is_nova_embed
            or self._inferred_provider == "cohere"
            or self._is_marengo
        )

    def _supports_audio(self) -> bool:
        """Check if the model supports audio embeddings (sync API)."""
        return self._is_nova_embed

    def _supports_video(self) -> bool:
        """Check if the model supports video embeddings (sync API)."""
        return self._is_nova_embed

    def _supports_s3_input(self) -> bool:
        """Check if the model supports S3 URI input."""
        return self._is_nova_embed or self._is_marengo

    def _get_dimensions_params(self) -> Dict[str, Any]:
        """Get dimensions parameter with provider-specific key name."""
        if self.dimensions is None:
            return {}

        if self._is_cohere_v4:
            return {"output_dimension": self.dimensions}
        elif self._inferred_provider == "cohere":
            return {}
        else:
            return {"dimensions": self.dimensions}

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate that AWS credentials to and python package exists in environment."""
        if self.client is None:
            self.client = create_aws_client(
                region_name=self.region_name,
                credentials_profile_name=self.credentials_profile_name,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
                endpoint_url=self.endpoint_url,
                config=self.config,
                service_name="bedrock-runtime",
            )

        return self

    def _embedding_func(
        self, text: str, input_type: str = "search_document"
    ) -> List[float]:
        """Call out to Bedrock embedding endpoint with a single text."""
        # replace newlines, which can negatively affect performance.
        text = text.replace(os.linesep, " ")

        if self._inferred_provider == "cohere":
            # Cohere input_type depends on usage
            # for embedding documents use "search_document"
            # for embedding queries for retrieval use "search_query"
            response_body = self._invoke_model(
                input_body={
                    "input_type": input_type,
                    "texts": [text],
                    **self._get_dimensions_params(),
                }
            )
            embeddings = response_body.get("embeddings")
            if embeddings is None:
                raise ValueError("No embeddings returned from model")
            # Embed v3 and v4 schemas
            if isinstance(embeddings, dict) and "float" in embeddings:
                processed_embeddings = embeddings["float"]
            else:
                processed_embeddings = embeddings
            return processed_embeddings[0]
        elif self._is_nova_embed:
            single_embedding_params: Dict[str, Any] = {
                "embeddingPurpose": "GENERIC_INDEX",
                "text": {
                    "truncationMode": "END",
                    "value": text,
                },
            }
            if self.dimensions:
                single_embedding_params["embeddingDimension"] = self.dimensions
            response_body = self._invoke_model(
                input_body={
                    "taskType": "SINGLE_EMBEDDING",
                    "singleEmbeddingParams": single_embedding_params,
                }
            )
            embeddings = response_body.get("embeddings")
            if not embeddings or not embeddings[0].get("embedding"):
                raise ValueError("No embedding returned from model")
            return embeddings[0]["embedding"]
        else:
            # includes common provider == "amazon"
            response_body = self._invoke_model(
                input_body={"inputText": text, **self._get_dimensions_params()},
            )
            embedding = response_body.get("embedding")
            if embedding is None:
                raise ValueError("No embedding returned from model")
            return embedding

    def _cohere_multi_embedding(self, texts: List[str]) -> List[List[float]]:
        """Call out to Cohere Bedrock embedding endpoint with multiple inputs."""
        # replace newlines, which can negatively affect performance.
        texts = [text.replace(os.linesep, " ") for text in texts]
        results: List[List[float]] = []

        # Iterate through the list of strings in batches
        for text_batch in _batch_cohere_embedding_texts(
            texts, is_v4=self._is_cohere_v4
        ):
            batch_embeddings = self._invoke_model(
                input_body={
                    "input_type": "search_document",
                    "texts": text_batch,
                    **self._get_dimensions_params(),
                }
            ).get("embeddings")
            # Embed v3 and v4 schemas
            if isinstance(batch_embeddings, dict) and "float" in batch_embeddings:
                processed_embeddings = batch_embeddings["float"]
            else:
                processed_embeddings = batch_embeddings

            if processed_embeddings is not None:
                results += processed_embeddings

        return results

    def _invoke_model(self, input_body: Dict[str, Any] = {}) -> Dict[str, Any]:
        if self.model_kwargs:
            input_body = {**input_body, **self.model_kwargs}

        body = json.dumps(input_body)

        try:
            response = self.client.invoke_model(
                body=body,
                modelId=self.model_id,
                accept="application/json",
                contentType="application/json",
            )

            response_metadata = response.get("ResponseMetadata", {})
            if response_metadata:
                logger.info(
                    f"Successfully invoked model {self.model_id}. "
                    f"ResponseMetadata: {response_metadata}"
                )

            response_body = json.loads(response.get("body").read())
            return response_body
        except Exception as e:
            logger.exception("Error raised by inference endpoint")
            raise e

    def _detect_media_format(self, data: bytes) -> str:
        """Detect media format from magic bytes."""
        if len(data) < 12:
            logger.warning(
                "Media data too short for format detection (%d bytes), "
                "defaulting to JPEG",
                len(data),
            )
            return "jpeg"

        # JPEG: FF D8 FF
        if data[:3] == b"\xff\xd8\xff":
            return "jpeg"
        # PNG: 89 50 4E 47 0D 0A 1A 0A
        if data[:8] == b"\x89PNG\r\n\x1a\n":
            return "png"
        # GIF: GIF87a or GIF89a
        if data[:6] in (b"GIF87a", b"GIF89a"):
            return "gif"
        # WEBP: RIFF....WEBP
        if data[:4] == b"RIFF" and data[8:12] == b"WEBP":
            return "webp"
        # MP4/MOV: ....ftyp
        if data[4:8] == b"ftyp":
            return "mp4"
        # MP3: FF FB or ID3
        if data[:2] == b"\xff\xfb" or data[:3] == b"ID3":
            return "mp3"
        # WAV: RIFF....WAVE
        if data[:4] == b"RIFF" and data[8:12] == b"WAVE":
            return "wav"
        # OGG: OggS
        if data[:4] == b"OggS":
            return "ogg"

        logger.warning(
            "Could not detect media format from magic bytes, defaulting to JPEG"
        )
        return "jpeg"

    def _load_media(
        self,
        data: MediaInput,
        media_type: Literal["image", "audio", "video"],
    ) -> tuple[str, str, Literal["inline", "s3"]]:
        """Normalize media input to (payload, format, source_kind).

        Args:
            data: Base64 string, file path, S3 URI, or raw bytes.
            media_type: Type of media (image, audio, video).

        Returns:
            Tuple of (payload, format, source_kind) where:
            - payload: base64 string or S3 URI
            - format: detected format (jpeg, png, mp3, etc.)
            - source_kind: "inline" or "s3"
        """
        if isinstance(data, bytes):
            fmt = self._detect_media_format(data)
            b64 = base64.b64encode(data).decode("utf-8")
            return b64, fmt, "inline"

        if data.startswith("s3://"):
            if not self._supports_s3_input():
                msg = f"S3 URIs not supported for model {self.model_id}"
                raise ValueError(msg)
            ext_match = re.search(r"\.([a-zA-Z0-9]+)$", data)
            fmt = ext_match.group(1).lower() if ext_match else "jpeg"
            if fmt == "jpg":
                fmt = "jpeg"
            return data, fmt, "s3"

        data_uri_match = re.match(
            r"^data:(?:image|video|audio)/([a-zA-Z0-9]+);base64,(.+)$", data
        )
        if data_uri_match:
            fmt = data_uri_match.group(1).lower()
            b64 = data_uri_match.group(2)
            return b64, fmt, "inline"

        if os.path.isfile(data):
            with open(data, "rb") as f:
                file_bytes = f.read()
            fmt = self._detect_media_format(file_bytes)
            b64 = base64.b64encode(file_bytes).decode("utf-8")
            return b64, fmt, "inline"

        # Assume raw base64 string
        try:
            decoded = base64.b64decode(data)
            fmt = self._detect_media_format(decoded)
            return data, fmt, "inline"
        except Exception:
            msg = "Could not interpret input as base64, file path, or S3 URI"
            raise ValueError(msg)

    def _build_image_request(
        self, payload: str, fmt: str, source_kind: str
    ) -> Dict[str, Any]:
        """Build provider-specific request body for image embedding."""
        if self._is_titan_multimodal:
            body: Dict[str, Any] = {"inputImage": payload}
            if self.dimensions:
                body["embeddingConfig"] = {"outputEmbeddingLength": self.dimensions}
            return body

        if self._is_nova_embed:
            params: Dict[str, Any] = {"embeddingPurpose": "GENERIC_INDEX"}
            if source_kind == "s3":
                params["image"] = {
                    "format": fmt,
                    "source": {"s3Location": {"uri": payload}},
                }
            else:
                params["image"] = {"format": fmt, "source": {"bytes": payload}}
            if self.dimensions:
                params["embeddingDimension"] = self.dimensions
            return {"taskType": "SINGLE_EMBEDDING", "singleEmbeddingParams": params}

        if self._inferred_provider == "cohere":
            data_uri = f"data:image/{fmt};base64,{payload}"
            if self._is_cohere_v4:
                return {
                    "input_type": "search_document",
                    "images": [data_uri],
                    **self._get_dimensions_params(),
                }
            return {
                "input_type": "image",
                "images": [data_uri],
                "embedding_types": ["float"],
            }

        if self._is_marengo:
            media_source: Dict[str, Any]
            if source_kind == "s3":
                media_source = {"s3Location": {"uri": payload}}
            else:
                media_source = {"base64String": payload}
            return {"inputType": "image", "image": {"mediaSource": media_source}}

        msg = f"Image embeddings not supported for model {self.model_id}"
        raise ValueError(msg)

    def _build_audio_request(
        self, payload: str, fmt: str, source_kind: str
    ) -> Dict[str, Any]:
        """Build request body for audio embedding (Nova only)."""
        params: Dict[str, Any] = {"embeddingPurpose": "GENERIC_INDEX"}
        if source_kind == "s3":
            params["audio"] = {
                "format": fmt,
                "source": {"s3Location": {"uri": payload}},
            }
        else:
            params["audio"] = {"format": fmt, "source": {"bytes": payload}}
        if self.dimensions:
            params["embeddingDimension"] = self.dimensions
        return {"taskType": "SINGLE_EMBEDDING", "singleEmbeddingParams": params}

    def _build_video_request(
        self, payload: str, fmt: str, source_kind: str
    ) -> Dict[str, Any]:
        """Build request body for video embedding (Nova only)."""
        params: Dict[str, Any] = {"embeddingPurpose": "GENERIC_INDEX"}
        if source_kind == "s3":
            params["video"] = {
                "format": fmt,
                "embeddingMode": "AUDIO_VIDEO_COMBINED",
                "source": {"s3Location": {"uri": payload}},
            }
        else:
            params["video"] = {
                "format": fmt,
                "embeddingMode": "AUDIO_VIDEO_COMBINED",
                "source": {"bytes": payload},
            }
        if self.dimensions:
            params["embeddingDimension"] = self.dimensions
        return {"taskType": "SINGLE_EMBEDDING", "singleEmbeddingParams": params}

    def _extract_media_embedding(self, response: Dict[str, Any]) -> List[float]:
        """Extract embedding from provider-specific response."""
        # Titan format
        if "embedding" in response and isinstance(response["embedding"], list):
            return response["embedding"]

        # Nova format
        if "embeddings" in response and isinstance(response["embeddings"], list):
            if response["embeddings"] and isinstance(response["embeddings"][0], dict):
                embedding = response["embeddings"][0].get("embedding")
                if embedding:
                    return embedding

        # Cohere format
        if "embeddings" in response:
            embeddings = response["embeddings"]
            if isinstance(embeddings, dict) and "float" in embeddings:
                return embeddings["float"][0]
            if isinstance(embeddings, list) and embeddings:
                return embeddings[0]

        # Marengo format: {"data": [{"embedding": [...]}]}
        if "data" in response and isinstance(response["data"], list):
            if response["data"] and isinstance(response["data"][0], dict):
                embedding = response["data"][0].get("embedding")
                if embedding:
                    return embedding

        msg = f"No embedding found in response. Keys: {list(response.keys())}"
        raise ValueError(msg)

    def _normalize_vector(self, embeddings: List[float]) -> List[float]:
        """Normalize the embedding to a unit vector."""
        emb = np.array(embeddings)
        norm_emb = emb / np.linalg.norm(emb)
        return norm_emb.tolist()

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Compute doc embeddings using a Bedrock model.

        Args:
            texts: The list of texts to embed

        Returns:
            List of embeddings, one for each text.

        """

        # If we are able to make use of Cohere's multiple embeddings, use that
        if self._inferred_provider == "cohere":
            return self._embed_cohere_documents(texts)
        else:
            return self._iteratively_embed_documents(texts)

    def _embed_cohere_documents(self, texts: List[str]) -> List[List[float]]:
        response = self._cohere_multi_embedding(texts)

        if self.normalize:
            response = [self._normalize_vector(embedding) for embedding in response]

        return response

    def _iteratively_embed_documents(self, texts: List[str]) -> List[List[float]]:
        results = []
        for text in texts:
            response = self._embedding_func(text)

            if self.normalize:
                response = self._normalize_vector(response)

            results.append(response)

        return results

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a Bedrock model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.

        """
        if self._inferred_provider == "cohere":
            embedding = self._embedding_func(text, input_type="search_query")
        else:
            embedding = self._embedding_func(text)

        if self.normalize:
            return self._normalize_vector(embedding)

        return embedding

    async def aembed_query(self, text: str) -> List[float]:
        """Asynchronous compute query embeddings using a Bedrock model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.

        """

        return await run_in_executor(None, self.embed_query, text)

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        """Asynchronous compute doc embeddings using a Bedrock model.

        Args:
            texts: The list of texts to embed

        Returns:
            List of embeddings, one for each text.

        """
        if self._inferred_provider == "cohere":
            return await run_in_executor(None, self.embed_documents, texts)

        result = await asyncio.gather(*[self.aembed_query(text) for text in texts])

        return list(result)

    def embed_image(self, image: MediaInput) -> List[float]:
        """Embed a single image.

        Args:
            image: Image as base64 string, file path, S3 URI (Nova/Marengo), or bytes.

        Returns:
            Embedding vector for the image.

        Raises:
            ValueError: If model doesn't support image embeddings.
        """
        if not self._supports_image():
            msg = f"Image embeddings not supported for model {self.model_id}"
            raise ValueError(msg)

        payload, fmt, source_kind = self._load_media(image, "image")
        body = self._build_image_request(payload, fmt, source_kind)
        response = self._invoke_model(body)
        embedding = self._extract_media_embedding(response)

        if self.normalize:
            return self._normalize_vector(embedding)
        return embedding

    def embed_images(self, images: List[MediaInput]) -> List[List[float]]:
        """Embed multiple images.

        Args:
            images: List of images as base64 strings, file paths, S3 URIs, or bytes.

        Returns:
            List of embedding vectors, one per image.
        """
        return [self.embed_image(img) for img in images]

    def embed_audio(self, audio: MediaInput) -> List[float]:
        """Embed audio content.

        Only supported by Nova models. Audio must be ≤30 seconds for sync API.

        Args:
            audio: Audio as base64 string, file path, S3 URI, or bytes.

        Returns:
            Embedding vector for the audio.

        Raises:
            ValueError: If model doesn't support audio embeddings.
        """
        if not self._supports_audio():
            msg = (
                f"Audio embeddings not supported for model {self.model_id}. "
                "Only Nova models support sync audio embeddings."
            )
            raise ValueError(msg)

        payload, fmt, source_kind = self._load_media(audio, "audio")
        body = self._build_audio_request(payload, fmt, source_kind)
        response = self._invoke_model(body)
        embedding = self._extract_media_embedding(response)

        if self.normalize:
            return self._normalize_vector(embedding)
        return embedding

    def embed_video(self, video: MediaInput) -> List[float]:
        """Embed video content.

        Only supported by Nova models. Video must be ≤30 seconds for sync API.

        Args:
            video: Video as base64 string, file path, S3 URI, or bytes.

        Returns:
            Embedding vector for the video.

        Raises:
            ValueError: If model doesn't support video embeddings.
        """
        if not self._supports_video():
            msg = (
                f"Video embeddings not supported for model {self.model_id}. "
                "Only Nova models support sync video embeddings."
            )
            raise ValueError(msg)

        payload, fmt, source_kind = self._load_media(video, "video")
        body = self._build_video_request(payload, fmt, source_kind)
        response = self._invoke_model(body)
        embedding = self._extract_media_embedding(response)

        if self.normalize:
            return self._normalize_vector(embedding)
        return embedding

    async def aembed_image(self, image: MediaInput) -> List[float]:
        """Asynchronously embed a single image."""
        return await run_in_executor(None, self.embed_image, image)

    async def aembed_images(self, images: List[MediaInput]) -> List[List[float]]:
        """Asynchronously embed multiple images."""
        result = await asyncio.gather(*[self.aembed_image(img) for img in images])
        return list(result)

    async def aembed_audio(self, audio: MediaInput) -> List[float]:
        """Asynchronously embed audio content."""
        return await run_in_executor(None, self.embed_audio, audio)

    async def aembed_video(self, video: MediaInput) -> List[float]:
        """Asynchronously embed video content."""
        return await run_in_executor(None, self.embed_video, video)


def _batch_cohere_embedding_texts(
    texts: List[str], is_v4: bool = False
) -> Generator[List[str], None, None]:
    """Batches a set of texts into chunks acceptable for the Cohere embedding API.

    Cohere allows up to 96 texts per request. Each individual text can be up to:
    - Embed v3: 2048 characters
    - Embed v4: ~128K tokens (~512,000 characters)

    The character limit applies per-text, not as a cumulative batch total.
    """
    max_items = 96
    if is_v4:
        # Cohere Embed v4 supports up to 128K tokens per input
        # Using conservative estimate of ~4 chars per token = ~512K chars
        # See: https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-embed-v4.html
        max_chars = 512_000
        char_limit_msg = (
            "The Cohere Embed v4 embedding API does not support texts longer than "
            "approximately 128K tokens (~512,000 characters)."
        )
    else:
        # Cohere Embed v3 limit
        # See: https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-embed.html
        max_chars = 2048
        char_limit_msg = (
            "The Cohere embedding API does not support texts longer than "
            "2048 characters."
        )

    # Validate all text lengths before yielding any batches
    if any(len(text) > max_chars for text in texts):
        raise ValueError(char_limit_msg)

    for i in range(0, len(texts), max_items):
        yield texts[i : i + max_items]



from langchain_aws.retrievers.bedrock import AmazonKnowledgeBasesRetriever
from langchain_aws.retrievers.kendra import AmazonKendraRetriever
from langchain_aws.retrievers.s3_vectors import AmazonS3VectorsRetriever

__all__ = [
    "AmazonKendraRetriever",
    "AmazonKnowledgeBasesRetriever",
    "AmazonS3VectorsRetriever",
]



import re
from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Union,
)

from langchain_core.callbacks import CallbackManagerForRetrieverRun
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever
from langchain_core.utils import secret_from_env
from pydantic import (
    BaseModel,
    Field,
    SecretStr,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated

from langchain_aws.utils import create_aws_client


def clean_excerpt(excerpt: str) -> str:
    """Clean an excerpt from Kendra.

    Args:
        excerpt: The excerpt to clean.

    Returns:
        The cleaned excerpt.

    """
    if not excerpt:
        return excerpt
    res = re.sub(r"\s+", " ", excerpt).replace("...", "")
    return res


def combined_text(item: "ResultItem") -> str:
    """Combine a ResultItem title and excerpt into a single string.

    Args:
        item: the ResultItem of a Kendra search.

    Returns:
        A combined text of the title and excerpt of the given item.

    """
    text = ""
    title = item.get_title()
    if title:
        text += f"Document Title: {title}\n"
    excerpt = clean_excerpt(item.get_excerpt())
    if excerpt:
        text += f"Document Excerpt: \n{excerpt}\n"
    return text


DocumentAttributeValueType = Union[str, int, List[str], None]
"""Possible types of a DocumentAttributeValue.

Dates are also represented as str.

"""


# Unexpected keyword argument "extra" for "__init_subclass__" of "object"
class Highlight(BaseModel, extra="allow"):  # type: ignore[call-arg]
    """Information that highlights the keywords in the excerpt."""

    BeginOffset: int
    """The zero-based location in the excerpt where the highlight starts."""
    EndOffset: int
    """The zero-based location in the excerpt where the highlight ends."""
    TopAnswer: Optional[bool] = None
    """Indicates whether the result is the best one."""
    Type: Optional[str] = None
    """The highlight type: STANDARD or THESAURUS_SYNONYM."""


# Unexpected keyword argument "extra" for "__init_subclass__" of "object"
class TextWithHighLights(BaseModel, extra="allow"):  # type: ignore[call-arg]
    """Text with highlights."""

    Text: str
    """The text."""
    Highlights: Optional[Any] = None
    """The highlights."""


# Unexpected keyword argument "extra" for "__init_subclass__" of "object"
class AdditionalResultAttributeValue(  # type: ignore[call-arg]
    BaseModel, extra="allow"
):
    """Value of an additional result attribute."""

    TextWithHighlightsValue: TextWithHighLights
    """The text with highlights value."""


# Unexpected keyword argument "extra" for "__init_subclass__" of "object"
class AdditionalResultAttribute(BaseModel, extra="allow"):  # type: ignore[call-arg]
    """Additional result attribute."""

    Key: str
    """The key of the attribute."""
    ValueType: Literal["TEXT_WITH_HIGHLIGHTS_VALUE"]
    """The type of the value."""
    Value: AdditionalResultAttributeValue
    """The value of the attribute."""

    def get_value_text(self) -> str:
        return self.Value.TextWithHighlightsValue.Text


# Unexpected keyword argument "extra" for "__init_subclass__" of "object"
class DocumentAttributeValue(BaseModel, extra="allow"):  # type: ignore[call-arg]
    """Value of a document attribute."""

    DateValue: Optional[str] = None
    """The date expressed as an ISO 8601 string."""
    LongValue: Optional[int] = None
    """The long value."""
    StringListValue: Optional[List[str]] = None
    """The string list value."""
    StringValue: Optional[str] = None
    """The string value."""

    @property
    def value(self) -> DocumentAttributeValueType:
        """The only defined document attribute value or None.

        According to Amazon Kendra, you can only provide one value for a document
        attribute.

        """
        if self.DateValue:
            return self.DateValue
        if self.LongValue:
            return self.LongValue
        if self.StringListValue:
            return self.StringListValue
        if self.StringValue:
            return self.StringValue

        return None


# Unexpected keyword argument "extra" for "__init_subclass__" of "object"
class DocumentAttribute(BaseModel, extra="allow"):  # type: ignore[call-arg]
    """Document attribute."""

    Key: str
    """The key of the attribute."""
    Value: DocumentAttributeValue
    """The value of the attribute."""


# Unexpected keyword argument "extra" for "__init_subclass__" of "object"
class ResultItem(BaseModel, ABC, extra="allow"):  # type: ignore[call-arg]
    """Base class of a result item."""

    Id: Optional[str] = None
    """The ID of the relevant result item."""
    DocumentId: Optional[str] = None
    """The document ID."""
    DocumentURI: Optional[str] = None
    """The document URI."""
    DocumentAttributes: Optional[List[DocumentAttribute]] = []
    """The document attributes."""
    ScoreAttributes: Optional[dict] = None
    """The kendra score confidence"""

    @abstractmethod
    def get_title(self) -> str:
        """Document title."""

    @abstractmethod
    def get_excerpt(self) -> str:
        """Document excerpt or passage original content as retrieved by Kendra."""

    def get_additional_metadata(self) -> dict:
        """Document additional metadata dict.
        This returns any extra metadata except these:
            * result_id
            * document_id
            * source
            * title
            * excerpt
            * document_attributes

        """
        return {}

    def get_document_attributes_dict(self) -> Dict[str, DocumentAttributeValueType]:
        """Document attributes dict."""
        return {attr.Key: attr.Value.value for attr in (self.DocumentAttributes or [])}

    def get_score_attribute(self) -> str:
        """Document Score Confidence"""
        if self.ScoreAttributes is not None:
            return self.ScoreAttributes["ScoreConfidence"]
        else:
            return "NOT_AVAILABLE"

    def to_doc(
        self, page_content_formatter: Callable[["ResultItem"], str] = combined_text
    ) -> Document:
        """Converts this item to a Document."""
        page_content = page_content_formatter(self)
        metadata = self.get_additional_metadata()
        metadata.update(
            {
                "result_id": self.Id,
                "document_id": self.DocumentId,
                "source": self.DocumentURI,
                "title": self.get_title(),
                "excerpt": self.get_excerpt(),
                "document_attributes": self.get_document_attributes_dict(),
                "score": self.get_score_attribute(),
            }
        )
        return Document(page_content=page_content, metadata=metadata)


class QueryResultItem(ResultItem):
    """Query API result item."""

    DocumentTitle: TextWithHighLights
    """The document title."""
    FeedbackToken: Optional[str] = None
    """Identifies a particular result from a particular query."""
    Format: Optional[str] = None
    """If the Type is ANSWER, then format is either:
        * TABLE: a table excerpt is returned in TableExcerpt;
        * TEXT: a text excerpt is returned in DocumentExcerpt.

    """
    Type: Optional[str] = None
    """Type of result: DOCUMENT or QUESTION_ANSWER or ANSWER"""
    AdditionalAttributes: Optional[List[AdditionalResultAttribute]] = []
    """One or more additional attributes associated with the result."""
    DocumentExcerpt: Optional[TextWithHighLights] = None
    """Excerpt of the document text."""

    def get_title(self) -> str:
        return self.DocumentTitle.Text

    def get_attribute_value(self) -> str:
        if not self.AdditionalAttributes:
            return ""
        if not self.AdditionalAttributes[0]:
            return ""
        else:
            return self.AdditionalAttributes[0].get_value_text()

    def get_excerpt(self) -> str:
        if (
            self.AdditionalAttributes
            and self.AdditionalAttributes[0].Key == "AnswerText"
        ):
            excerpt = self.get_attribute_value()
        elif self.DocumentExcerpt:
            excerpt = self.DocumentExcerpt.Text
        else:
            excerpt = ""

        return excerpt

    def get_additional_metadata(self) -> dict:
        additional_metadata = {"type": self.Type}
        return additional_metadata


class RetrieveResultItem(ResultItem):
    """Retrieve API result item."""

    DocumentTitle: Optional[str] = None
    """The document title."""
    Content: Optional[str] = None
    """The content of the item."""

    def get_title(self) -> str:
        return self.DocumentTitle or ""

    def get_excerpt(self) -> str:
        return self.Content or ""


# Unexpected keyword argument "extra" for "__init_subclass__" of "object"
class QueryResult(BaseModel, extra="allow"):  # type: ignore[call-arg]
    """`Amazon Kendra Query API` search result.

    It is composed of:
        * Relevant suggested answers: either a text excerpt or table excerpt.
        * Matching FAQs or questions-answer from your FAQ file.
        * Documents including an excerpt of each document with its title.

    """

    ResultItems: List[QueryResultItem]
    """The result items."""


# Unexpected keyword argument "extra" for "__init_subclass__" of "object"
class RetrieveResult(BaseModel, extra="allow"):  # type: ignore[call-arg]
    """`Amazon Kendra Retrieve API` search result.

    It is composed of:
        * relevant passages or text excerpts given an input query.

    """

    QueryId: str
    """The ID of the query."""
    ResultItems: List[RetrieveResultItem]
    """The result items."""


KENDRA_CONFIDENCE_MAPPING = {
    "NOT_AVAILABLE": 0.0,
    "LOW": 0.25,
    "MEDIUM": 0.50,
    "HIGH": 0.75,
    "VERY_HIGH": 1.0,
}


class AmazonKendraRetriever(BaseRetriever):
    """`Amazon Kendra Index` retriever.

    Attributes:
        index_id: Kendra index id

        region_name: The aws region e.g., `'us-west-2'`.
            Falls back to `AWS_REGION`/`AWS_DEFAULT_REGION` env variable
            or region specified in `~/.aws/config`.

        credentials_profile_name: The name of the profile in the `~/.aws/credentials`
            or `~/.aws/config` files, which has either access keys or role information
            specified. If not specified, the default credential profile or, if on an
            EC2 instance, credentials from IMDS will be used.

        aws_access_key_id: AWS access key id. If provided, ``aws_secret_access_key``
            must also be provided. If not specified, the default credential profile or,
            if on an EC2 instance, credentials from IMDS will be used.

            See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html

            If not provided, will be read from `AWS_ACCESS_KEY_ID` environment
            variable.

        aws_secret_access_key: AWS secret_access_key. If provided, `aws_access_key_id`
            must also be provided. If not specified, the default credential profile or,
            if on an EC2 instance, credentials from IMDS will be used.

            See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html

            If not provided, will be read from `AWS_SECRET_ACCESS_KEY` environment
            variable.

        aws_session_token: AWS session token. If provided, `aws_access_key_id` and
            `aws_secret_access_key` must also be provided. Not required unless using
            temporary credentials.

            See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html

            If not provided, will be read from `AWS_SESSION_TOKEN` environment
            variable.

        endpoint_url: Needed if you don't want to default to `'us-east-1'` endpoint.

        config: An optional `botocore.config.Config` instance to pass to the client.

        top_k: No of results to return

        attribute_filter: Additional filtering of results based on metadata

            See: https://docs.aws.amazon.com/kendra/latest/APIReference

        page_content_formatter: generates the Document page_content
            allowing access to all result item attributes. By default, it uses
            the item's title and excerpt.

        client: boto3 client for Kendra

        user_context: Provides information about the user context
            See: https://docs.aws.amazon.com/kendra/latest/APIReference

    Example:
        ```python
        retriever = AmazonKendraRetriever(
            index_id="c0806df7-e76b-4bce-9b5c-d5582f6b1a03"
        )
        ```

    """

    index_id: str
    region_name: Optional[str] = None
    credentials_profile_name: Optional[str] = None
    aws_access_key_id: Optional[SecretStr] = Field(
        default_factory=secret_from_env("AWS_ACCESS_KEY_ID", default=None)
    )
    aws_secret_access_key: Optional[SecretStr] = Field(
        default_factory=secret_from_env("AWS_SECRET_ACCESS_KEY", default=None)
    )
    aws_session_token: Optional[SecretStr] = Field(
        default_factory=secret_from_env("AWS_SESSION_TOKEN", default=None)
    )
    endpoint_url: Optional[str] = None
    config: Any = None
    client: Any = None
    top_k: int = 3
    attribute_filter: Optional[Dict] = None
    page_content_formatter: Callable[[ResultItem], str] = combined_text
    user_context: Optional[Dict] = None
    min_score_confidence: Annotated[Optional[float], Field(ge=0.0, le=1.0)]

    @field_validator("top_k")
    def validate_top_k(cls, value: int) -> int:
        if value < 0:
            raise ValueError(f"top_k ({value}) cannot be negative.")
        return value

    @model_validator(mode="before")
    @classmethod
    def create_client(cls, values: Dict[str, Any]) -> Any:
        if values.get("client") is None:
            values["client"] = create_aws_client(
                region_name=values.get("region_name"),
                credentials_profile_name=values.get("credentials_profile_name"),
                aws_access_key_id=values.get("aws_access_key_id"),
                aws_secret_access_key=values.get("aws_secret_access_key"),
                aws_session_token=values.get("aws_session_token"),
                endpoint_url=values.get("endpoint_url"),
                config=values.get("config"),
                service_name="kendra",
            )

        return values

    def _kendra_query(self, query: str) -> Sequence[ResultItem]:
        kendra_kwargs = {
            "IndexId": self.index_id,
            # truncate the query to ensure that
            # there is no validation exception from Kendra.
            "QueryText": query.strip()[0:999],
            "PageSize": self.top_k,
        }
        if self.attribute_filter is not None:
            kendra_kwargs["AttributeFilter"] = self.attribute_filter
        if self.user_context is not None:
            kendra_kwargs["UserContext"] = self.user_context

        response = self.client.retrieve(**kendra_kwargs)
        r_result = RetrieveResult.model_validate(response)
        if r_result.ResultItems:
            return r_result.ResultItems

        # Retrieve API returned 0 results, fall back to Query API
        response = self.client.query(**kendra_kwargs)
        q_result = QueryResult.model_validate(response)
        return q_result.ResultItems

    def _get_top_k_docs(self, result_items: Sequence[ResultItem]) -> List[Document]:
        top_docs = [
            item.to_doc(self.page_content_formatter)
            for item in result_items[: self.top_k]
        ]
        return top_docs

    def _filter_by_score_confidence(self, docs: List[Document]) -> List[Document]:
        """Filter out the records that have a score confidence less than the required
        threshold.

        """
        if not self.min_score_confidence:
            return docs
        filtered_docs = [
            item
            for item in docs
            if (
                item.metadata.get("score") is not None
                and isinstance(item.metadata["score"], str)
                and KENDRA_CONFIDENCE_MAPPING.get(item.metadata["score"], 0.0)
                >= self.min_score_confidence
            )
        ]
        return filtered_docs

    def _get_relevant_documents(
        self,
        query: str,
        *,
        run_manager: CallbackManagerForRetrieverRun,
    ) -> List[Document]:
        """Run search on Kendra index and get top k documents

        Example:
            ```python
            docs = retriever.get_relevant_documents('This is my query')
            ```
        """
        result_items = self._kendra_query(query)
        top_k_docs = self._get_top_k_docs(result_items)
        return self._filter_by_score_confidence(top_k_docs)



# type:ignore

import json
from typing import AsyncGenerator, Dict
from unittest.mock import MagicMock, patch

import pytest
from langchain_core.messages import AIMessageChunk

from langchain_aws import BedrockLLM
from langchain_aws.llms.bedrock import (
    ALTERNATION_ERROR,
    LLMInputOutputAdapter,
    _human_assistant_format,
)

TEST_CASES = {
    """Hey""": """

Human: Hey

Assistant:""",
    """

Human: Hello

Assistant:""": """

Human: Hello

Assistant:""",
    """Human: Hello

Assistant:""": """

Human: Hello

Assistant:""",
    """
Human: Hello

Assistant:""": """

Human: Hello

Assistant:""",
    """

Human: Human: Hello

Assistant:""": (
        "Error: Prompt must alternate between '\n\nHuman:' and '\n\nAssistant:'."
    ),
    """Human: Hello

Assistant: Hello

Human: Hello

Assistant:""": """

Human: Hello

Assistant: Hello

Human: Hello

Assistant:""",
    """

Human: Hello

Assistant: Hello

Human: Hello

Assistant:""": """

Human: Hello

Assistant: Hello

Human: Hello

Assistant:""",
    """

Human: Hello

Assistant: Hello

Human: Hello

Assistant: Hello

Assistant: Hello""": ALTERNATION_ERROR,
    """

Human: Hi.

Assistant: Hi.

Human: Hi.

Human: Hi.

Assistant:""": ALTERNATION_ERROR,
    """
Human: Hello""": """

Human: Hello

Assistant:""",
    """

Human: Hello
Hello

Assistant""": """

Human: Hello
Hello

Assistant

Assistant:""",
    """Hello

Assistant:""": """

Human: Hello

Assistant:""",
    """Hello

Human: Hello

""": """Hello

Human: Hello



Assistant:""",
    """

Human: Assistant: Hello""": """

Human: 

Assistant: Hello""",
    """

Human: Human

Assistant: Assistant

Human: Assistant

Assistant: Human""": """

Human: Human

Assistant: Assistant

Human: Assistant

Assistant: Human""",
    """
Assistant: Hello there, your name is:

Human.

Human: Hello there, your name is: 

Assistant.""": """

Human: 

Assistant: Hello there, your name is:

Human.

Human: Hello there, your name is: 

Assistant.

Assistant:""",
    """

Human: Human: Hi

Assistant: Hi""": ALTERNATION_ERROR,
    """Human: Hi

Human: Hi""": ALTERNATION_ERROR,
    """

Assistant: Hi

Human: Hi""": """

Human: 

Assistant: Hi

Human: Hi

Assistant:""",
    """

Human: Hi

Assistant: Yo

Human: Hey

Assistant: Sup

Human: Hi

Assistant: Hi
Human: Hi
Assistant:""": """

Human: Hi

Assistant: Yo

Human: Hey

Assistant: Sup

Human: Hi

Assistant: Hi

Human: Hi

Assistant:""",
    """

Hello.

Human: Hello.

Assistant:""": """

Hello.

Human: Hello.

Assistant:""",
}


def test__human_assistant_format() -> None:
    for input_text, expected_output in TEST_CASES.items():
        if expected_output == ALTERNATION_ERROR:
            with pytest.warns(UserWarning, match=ALTERNATION_ERROR):
                _human_assistant_format(input_text)
        else:
            output = _human_assistant_format(input_text)
            assert output == expected_output


# Sample mock streaming response data
MOCK_STREAMING_RESPONSE = [
    {"chunk": {"bytes": b'{"text": "nice"}'}},
    {"chunk": {"bytes": b'{"text": " to meet"}'}},
    {"chunk": {"bytes": b'{"text": " you"}'}},
]

MOCK_STREAMING_RESPONSE_MISTRAL = [
    {"chunk": {"bytes": b'{"outputs": [{"text": "Thank","stop_reason": null}]}'}},
    {"chunk": {"bytes": b'{"outputs": [{"text": "you.","stop_reason": "stop"}]}'}},
]

MOCK_STREAMING_RESPONSE_DEEPSEEK = [
    {"chunk": {"bytes": b'{"choices": [{"text": "Thank","stop_reason": null}]}'}},
    {"chunk": {"bytes": b'{"choices": [{"text": "you.","stop_reason": "stop"}]}'}},
]

MOCK_STREAMING_RESPONSE_ANTHROPIC = [
    {"chunk": {"bytes": b'{"completion": "Thank", "stop_reason": null}'}},
    {"chunk": {"bytes": b'{"completion": " you.", "stop_reason": "stop_sequence"}'}},
]

MOCK_STREAMING_RESPONSE_WRITER = [
    {
        "chunk": {
            "bytes": b'{"id":"cmpl-ec61121fa19443caa7f614bde08e926c",'
            b'"object":"text_completion",'
            b'"created":1747106231,'
            b'"model":"writer.palmyra-x5-v1:0",'
            b'"choices":[{"index":0,"text":"Hel","logprobs":null,"finish_reason":null,"stop_reason":null}],'
            b'"usage":null}'
        }
    },
    {
        "chunk": {
            "bytes": b'{"id":"cmpl-ec61121fa19443caa7f614bde08e926c",'
            b'"object":"text_completion",'
            b'"created":1747106231,'
            b'"model":"writer.palmyra-x5-v1:0",'
            b'"choices":[{"index":0,"text":"lo.","logprobs":null,"finish_reason":"length","stop_reason":null}],'
            b'"usage":null}'
        }
    },
    {"chunk": {"bytes": b'"[DONE]"'}},
]

MOCK_STREAMING_RESPONSE_QWEN = [
    {
        "chunk": {
            "bytes": b'{"choices": [{"delta": {"content": "", "role": "assistant"}, '
            b'"finish_reason": null, "index": 0}], '
            b'"created": 1759875373, '
            b'"id": "chatcmpl-a069cbda08ce4599afae798c4d2de095", '
            b'"model": "qwen.qwen3-32b-v1:0", '
            b'"object": "chat.completion.chunk", '
            b'"service_tier": "auto"}'
        }
    },
    {
        "chunk": {
            "bytes": b'{"choices": [{"delta": {"content": "Hello.  \\nGoodbye."}, '
            b'"finish_reason": "stop", "index": 0}], '
            b'"created": 1759875373, '
            b'"id": "chatcmpl-a069cbda08ce4599afae798c4d2de095", '
            b'"model": "qwen.qwen3-32b-v1:0", '
            b'"object": "chat.completion.chunk", '
            b'"service_tier": "auto", '
            b'"amazon-bedrock-invocationMetrics": {'
            b'"inputTokenCount": 35, "outputTokenCount": 7, '
            b'"invocationLatency": 225, "firstByteLatency": 191}}'
        }
    },
]

MOCK_STREAMING_RESPONSE_OPENAI = [
    {
        "chunk": {
            "bytes": b'{"choices": [{"delta": {"content": "Hello."}, '
            b'"finish_reason": null, "index": 0}], '
            b'"created": 1759813667, '
            b'"id": "chatcmpl-fa6fb768b71046eeb3880cbb4a1b07c1", '
            b'"model": "openai.gpt-oss-20b-1:0", '
            b'"object": "chat.completion.chunk", "service_tier": "auto"}'
        }
    },
    {
        "chunk": {
            "bytes": b'{"choices": [{"delta": {}, '
            b'"finish_reason": "stop", "index": 0}],'
            b' "created": 1759813667, '
            b'"id": "chatcmpl-fa6fb768b71046eeb3880cbb4a1b07c1", '
            b'"model": "openai.gpt-oss-20b-1:0", '
            b'"object": "chat.completion.chunk", '
            b'"service_tier": "auto", '
            b'"amazon-bedrock-invocationMetrics": {'
            b'"inputTokenCount": 84, '
            b'"outputTokenCount": 87, '
            b'"invocationLatency": 3981, '
            b'"firstByteLatency": 3615}}'
        }
    },
]


async def async_gen_mock_streaming_response() -> AsyncGenerator[Dict, None]:
    for item in MOCK_STREAMING_RESPONSE:
        yield item


@pytest.mark.asyncio
async def test_bedrock_async_streaming_call() -> None:
    # Mock boto3 import
    mock_boto3 = MagicMock()
    mock_boto3.Session.return_value.client.return_value = (
        MagicMock()
    )  # Mocking the client method of the Session object

    with patch.dict(
        "sys.modules", {"boto3": mock_boto3}
    ):  # Mocking boto3 at the top level using patch.dict
        # Mock the `BedrockLLM` class's method that invokes the model
        mock_invoke_method = MagicMock(return_value=async_gen_mock_streaming_response())
        with patch.object(
            BedrockLLM, "_aprepare_input_and_invoke_stream", mock_invoke_method
        ):
            # Instantiate the Bedrock LLM
            llm = BedrockLLM(
                client=None,
                model_id="anthropic.claude-v2",
                streaming=True,
                region_name="us-west-2",
            )
            # Call the _astream method
            chunks = [
                json.loads(chunk["chunk"]["bytes"])["text"]  # type: ignore
                async for chunk in llm._astream("Hey, how are you?")
            ]

    # Assertions
    assert len(chunks) == 3
    assert chunks[0] == "nice"
    assert chunks[1] == " to meet"
    assert chunks[2] == " you"


@pytest.fixture
def mistral_response():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {"outputs": [{"text": "This is the Mistral output text."}]}
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "18",
                "x-amzn-bedrock-output-token-count": "28",
            }
        },
    )

    return response


@pytest.fixture
def mistral_streaming_response():
    response = dict(body=MOCK_STREAMING_RESPONSE_MISTRAL)
    return response


@pytest.fixture
def deepseek_response():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {"choices": [{"text": "This is the DeepSeek output text."}]}
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "41",
                "x-amzn-bedrock-output-token-count": "51",
            }
        },
    )

    return response


@pytest.fixture
def deepseek_streaming_response():
    response = dict(body=MOCK_STREAMING_RESPONSE_DEEPSEEK)
    return response


@pytest.fixture
def writer_response():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {"choices": [{"text": " This is the Writer output text."}]}
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "17",
                "x-amzn-bedrock-output-token-count": "8",
            }
        },
    )

    return response


@pytest.fixture
def writer_streaming_response():
    response = dict(body=MOCK_STREAMING_RESPONSE_WRITER)
    return response


@pytest.fixture
def qwen_response():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {"choices": [{"message": {"content": "This is the Qwen output text."}}]}
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "35",
                "x-amzn-bedrock-output-token-count": "42",
            }
        },
    )

    return response


@pytest.fixture
def qwen_streaming_response():
    response = dict(body=MOCK_STREAMING_RESPONSE_QWEN)
    return response


@pytest.fixture
def openai_response():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {"choices": [{"message": {"content": "This is the OpenAI output text."}}]}
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "85",
                "x-amzn-bedrock-output-token-count": "80",
            }
        },
    )

    return response


@pytest.fixture
def openai_streaming_response():
    response = dict(body=MOCK_STREAMING_RESPONSE_OPENAI)
    return response


@pytest.fixture
def cohere_response():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {"generations": [{"text": "This is the Cohere output text."}]}
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "12",
                "x-amzn-bedrock-output-token-count": "22",
            }
        },
    )
    return response


@pytest.fixture
def anthropic_response():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {"completion": "This is the output text."}
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "10",
                "x-amzn-bedrock-output-token-count": "20",
            }
        },
    )
    return response


@pytest.fixture
def ai21_response():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {"completions": [{"data": {"text": "This is the AI21 output text."}}]}
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "15",
                "x-amzn-bedrock-output-token-count": "25",
            }
        },
    )
    return response


@pytest.fixture
def response_with_stop_reason():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {"completion": "This is the output text.", "stop_reason": "length"}
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "10",
                "x-amzn-bedrock-output-token-count": "20",
            }
        },
    )
    return response


def test_prepare_output_for_mistral(mistral_response):
    result = LLMInputOutputAdapter.prepare_output("mistral", mistral_response)
    assert result["text"] == "This is the Mistral output text."
    assert result["usage"]["prompt_tokens"] == 18
    assert result["usage"]["completion_tokens"] == 28
    assert result["usage"]["total_tokens"] == 46
    assert result["stop_reason"] is None


def test_prepare_output_stream_for_mistral(mistral_streaming_response) -> None:
    results = [
        chunk.text
        for chunk in LLMInputOutputAdapter.prepare_output_stream(
            "mistral", mistral_streaming_response
        )
    ]

    assert results[0] == "Thank"
    assert results[1] == "you."


def test_prepare_output_for_deepseek(deepseek_response):
    result = LLMInputOutputAdapter.prepare_output("deepseek", deepseek_response)
    assert result["text"] == "This is the DeepSeek output text."
    assert result["usage"]["prompt_tokens"] == 41
    assert result["usage"]["completion_tokens"] == 51
    assert result["usage"]["total_tokens"] == 92
    assert result["stop_reason"] is None


def test_prepare_output_stream_for_deepseek(deepseek_streaming_response) -> None:
    results = [
        chunk.text
        for chunk in LLMInputOutputAdapter.prepare_output_stream(
            "deepseek", deepseek_streaming_response
        )
    ]

    assert results[0] == "Thank"
    assert results[1] == "you."


def test_prepare_output_for_writer(writer_response):
    result = LLMInputOutputAdapter.prepare_output("writer", writer_response)
    assert result["text"] == " This is the Writer output text."
    assert result["usage"]["prompt_tokens"] == 17
    assert result["usage"]["completion_tokens"] == 8
    assert result["usage"]["total_tokens"] == 25
    assert result["stop_reason"] is None


def test_prepare_output_stream_for_writer(writer_streaming_response) -> None:
    results = [
        chunk.text
        for chunk in LLMInputOutputAdapter.prepare_output_stream(
            "writer", writer_streaming_response
        )
    ]

    assert results[0] == "Hel"
    assert results[1] == "lo."


def test_prepare_output_for_qwen(qwen_response):
    result = LLMInputOutputAdapter.prepare_output("qwen", qwen_response)
    assert result["text"] == "This is the Qwen output text."
    assert result["usage"]["prompt_tokens"] == 35
    assert result["usage"]["completion_tokens"] == 42
    assert result["usage"]["total_tokens"] == 77
    assert result["stop_reason"] is None


def test_prepare_output_stream_for_qwen(qwen_streaming_response) -> None:
    results = [
        chunk.text
        for chunk in LLMInputOutputAdapter.prepare_output_stream(
            "qwen", qwen_streaming_response
        )
    ]

    assert results[0] == ""
    assert results[1] == "Hello.  \nGoodbye."


def test_prepare_output_for_openai(openai_response):
    result = LLMInputOutputAdapter.prepare_output("openai", openai_response)
    assert result["text"] == "This is the OpenAI output text."
    assert result["usage"]["prompt_tokens"] == 85
    assert result["usage"]["completion_tokens"] == 80
    assert result["usage"]["total_tokens"] == 165
    assert result["stop_reason"] is None


def test_prepare_output_stream_for_openai(openai_streaming_response) -> None:
    results = [
        chunk.text
        for chunk in LLMInputOutputAdapter.prepare_output_stream(
            "openai", openai_streaming_response
        )
    ]

    assert results[0] == "Hello."
    assert results[1] == ""


def test_prepare_output_for_cohere(cohere_response):
    result = LLMInputOutputAdapter.prepare_output("cohere", cohere_response)
    assert result["text"] == "This is the Cohere output text."
    assert result["usage"]["prompt_tokens"] == 12
    assert result["usage"]["completion_tokens"] == 22
    assert result["usage"]["total_tokens"] == 34
    assert result["stop_reason"] is None


def test_prepare_output_with_stop_reason(response_with_stop_reason):
    result = LLMInputOutputAdapter.prepare_output(
        "anthropic", response_with_stop_reason
    )
    assert result["text"] == "This is the output text."
    assert result["usage"]["prompt_tokens"] == 10
    assert result["usage"]["completion_tokens"] == 20
    assert result["usage"]["total_tokens"] == 30
    assert result["stop_reason"] == "length"


def test_prepare_output_for_anthropic(anthropic_response):
    result = LLMInputOutputAdapter.prepare_output("anthropic", anthropic_response)
    assert result["text"] == "This is the output text."
    assert result["usage"]["prompt_tokens"] == 10
    assert result["usage"]["completion_tokens"] == 20
    assert result["usage"]["total_tokens"] == 30
    assert result["stop_reason"] is None


def test_prepare_output_for_ai21(ai21_response):
    result = LLMInputOutputAdapter.prepare_output("ai21", ai21_response)
    assert result["text"] == "This is the AI21 output text."
    assert result["usage"]["prompt_tokens"] == 15
    assert result["usage"]["completion_tokens"] == 25
    assert result["usage"]["total_tokens"] == 40
    assert result["stop_reason"] is None


def test_standard_tracing_params():
    llm = BedrockLLM(model_id="foo", region_name="us-west-2")
    ls_params = llm._get_ls_params()
    assert ls_params == {
        "ls_provider": "amazon_bedrock",
        "ls_model_type": "llm",
        "ls_model_name": "foo",
    }


@pytest.fixture
def anthropic_response_with_thinking():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {
            "content": [
                {
                    "type": "thinking",
                    "thinking": "Let me think through this step by step...",
                    "signature": "SIGNATURE123",
                },
                {"type": "text", "text": "This is the output text."},
            ]
        }
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "10",
                "x-amzn-bedrock-output-token-count": "30",
            }
        },
    )
    return response


@pytest.fixture
def anthropic_response_with_thinking_and_tool_use():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {
            "content": [
                {
                    "type": "thinking",
                    "thinking": "I need to use a tool to answer this question...",
                    "signature": "SIGNATURE456",
                },
                {"type": "text", "text": "Let me check that for you."},
                {
                    "type": "tool_use",
                    "id": "tool_1",
                    "name": "get_weather",
                    "input": {"city": "nyc"},
                },
            ],
            "stop_reason": "tool_use",
        }
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "15",
                "x-amzn-bedrock-output-token-count": "40",
            }
        },
    )
    return response


@pytest.fixture
def anthropic_response_after_tool_use():
    body = MagicMock()
    body.read.return_value = json.dumps(
        {
            "content": [
                {"type": "text", "text": "Based on the data, it's cloudy in NYC."},
            ],
            "stop_reason": "end_turn",
        }
    ).encode()
    response = dict(
        body=body,
        ResponseMetadata={
            "HTTPHeaders": {
                "x-amzn-bedrock-input-token-count": "60",
                "x-amzn-bedrock-output-token-count": "20",
            }
        },
    )
    return response


def test_prepare_output_with_thinking(anthropic_response_with_thinking):
    """Test that thinking blocks are extracted properly from the response."""
    result = LLMInputOutputAdapter.prepare_output(
        "anthropic", anthropic_response_with_thinking
    )

    # Check that the text content was extracted correctly
    assert result["text"] == "This is the output text."

    # Check that the thinking block was extracted correctly
    assert "thinking" in result
    assert isinstance(result["thinking"], dict)
    assert result["thinking"]["text"] == "Let me think through this step by step..."
    assert result["thinking"]["signature"] == "SIGNATURE123"

    # Check that token counts are correct
    assert result["usage"]["prompt_tokens"] == 10
    assert result["usage"]["completion_tokens"] == 30
    assert result["usage"]["total_tokens"] == 40


def test_prepare_output_with_thinking_and_tool_use(
    anthropic_response_with_thinking_and_tool_use,
):
    """Test that thinking blocks and tool use are
    extracted properly from the response."""
    result = LLMInputOutputAdapter.prepare_output(
        "anthropic", anthropic_response_with_thinking_and_tool_use
    )

    # Check that the text content was extracted correctly
    assert result["text"] == "Let me check that for you."

    # Check that the thinking block was extracted correctly
    assert "thinking" in result
    assert isinstance(result["thinking"], dict)
    assert (
        result["thinking"]["text"] == "I need to use a tool to answer this question..."
    )
    assert result["thinking"]["signature"] == "SIGNATURE456"

    # Check that tool calls are extracted correctly
    assert "tool_calls" in result
    assert len(result["tool_calls"]) == 1
    assert result["tool_calls"][0]["name"] == "get_weather"
    assert result["tool_calls"][0]["args"] == {"city": "nyc"}
    assert result["tool_calls"][0]["id"] == "tool_1"

    # Check that stop reason is correctly extracted
    assert result["stop_reason"] == "tool_use"

    # Check that token counts are correct
    assert result["usage"]["prompt_tokens"] == 15
    assert result["usage"]["completion_tokens"] == 40
    assert result["usage"]["total_tokens"] == 55


def test_prepare_output_after_tool_use(anthropic_response_after_tool_use):
    """Test that responses after tool use (which don't have thinking blocks)
    are handled correctly."""
    result = LLMInputOutputAdapter.prepare_output(
        "anthropic", anthropic_response_after_tool_use
    )

    # Check that the text content was extracted correctly
    assert result["text"] == "Based on the data, it's cloudy in NYC."

    # Check that thinking is an empty dictionary when no thinking blocks are present
    assert "thinking" in result
    assert result["thinking"] == {}

    # Check that stop reason is correctly extracted
    assert result["stop_reason"] == "end_turn"

    # Check that token counts are correct
    assert result["usage"]["prompt_tokens"] == 60
    assert result["usage"]["completion_tokens"] == 20
    assert result["usage"]["total_tokens"] == 80


def test__get_base_model():
    """Test that _get_base_model returns the expected result."""
    llm = BedrockLLM(model_id="meta.llama3-8b-instruct-v1:0", region_name="us-west-2")
    assert llm._get_base_model() == "llama3-8b-instruct-v1:0"

    llm = BedrockLLM(
        model_id="arn:aws:bedrock:us-east-1::custom-model/meta.llama3-8b-instruct-v1:0/MyModel",
        base_model_id="meta.llama3-8b-instruct-v1:0",
        provider="meta",
        region_name="us-west-2",
    )
    assert llm._get_base_model() == "meta.llama3-8b-instruct-v1:0"

    llm = BedrockLLM(
        model_id="meta.llama2-70b-v1",
        base_model_id="meta.llama3-8b-instruct-v1:0",
        region_name="us-west-2",
    )
    assert llm._get_base_model() == "meta.llama3-8b-instruct-v1:0"


@patch("langchain_aws.llms.bedrock.create_aws_client")
def test_bedrock_client_creation(mock_create_client):
    """Test that both bedrock-runtime and bedrock clients are created."""
    mock_runtime_client = MagicMock()
    mock_bedrock_client = MagicMock()
    mock_create_client.side_effect = [mock_runtime_client, mock_bedrock_client]

    llm = BedrockLLM(model_id="meta.llama3-8b-instruct-v1:0", region_name="us-west-2")

    # Should create both clients
    assert mock_create_client.call_count == 2

    # Check that bedrock-runtime client was created
    calls = mock_create_client.call_args_list
    runtime_call = calls[0]
    assert runtime_call.kwargs["service_name"] == "bedrock-runtime"
    assert runtime_call.kwargs["region_name"] == "us-west-2"

    # Check that bedrock client was created
    bedrock_call = calls[1]
    assert bedrock_call.kwargs["service_name"] == "bedrock"
    assert bedrock_call.kwargs["region_name"] == "us-west-2"

    assert llm.client is mock_runtime_client
    assert llm.bedrock_client is mock_bedrock_client


@patch("langchain_aws.llms.bedrock.create_aws_client")
def test_get_base_model_with_application_inference_profile(mock_create_client):
    """Test _get_base_model with application inference profile."""
    mock_runtime_client = MagicMock()
    mock_bedrock_client = MagicMock()
    mock_bedrock_client.get_inference_profile.return_value = {
        "models": [
            {
                "modelArn": "arn:aws:bedrock:us-east-1::foundation-model/anthropic.claude-sonnet-4-20250514-v1:0"  # noqa: E501
            }
        ]
    }
    mock_create_client.side_effect = [mock_runtime_client, mock_bedrock_client]

    llm = BedrockLLM(
        model_id="arn:aws:bedrock:us-east-1:123456789012:application-inference-profile/my-profile",
        provider="anthropic",
        region_name="us-west-2",
    )

    result = llm._get_base_model()

    # Should call get_inference_profile and extract base model
    mock_bedrock_client.get_inference_profile.assert_called_once_with(
        inferenceProfileIdentifier="arn:aws:bedrock:us-east-1:123456789012:application-inference-profile/my-profile"
    )
    assert result == "anthropic.claude-sonnet-4-20250514-v1:0"
    assert llm.base_model_id == "anthropic.claude-sonnet-4-20250514-v1:0"


@pytest.fixture
def anthropic_streaming_response_with_close():
    """Streaming response with a mock body that tracks close() calls."""
    mock_body = MagicMock()
    mock_body.__iter__ = MagicMock(return_value=iter(MOCK_STREAMING_RESPONSE_ANTHROPIC))
    return {"body": mock_body}


def test_stream_closes_response_body(
    anthropic_streaming_response_with_close,
) -> None:
    """Test that streaming closes the response body after iteration."""
    mock_client = MagicMock()
    mock_client.invoke_model_with_response_stream.return_value = (
        anthropic_streaming_response_with_close
    )

    llm = BedrockLLM(
        client=mock_client,
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        region="us-west-2",
    )

    list(llm._prepare_input_and_invoke_stream(prompt="Hello"))

    anthropic_streaming_response_with_close["body"].close.assert_called_once()


def test_stream_closes_response_body_on_exception() -> None:
    """Test that streaming closes body even when iteration raises."""
    mock_client = MagicMock()
    mock_body = MagicMock()

    def raise_error():
        yield from MOCK_STREAMING_RESPONSE_ANTHROPIC
        raise RuntimeError("Test error")

    mock_body.__iter__ = MagicMock(return_value=raise_error())
    mock_client.invoke_model_with_response_stream.return_value = {"body": mock_body}

    llm = BedrockLLM(
        client=mock_client,
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        region="us-west-2",
    )

    with pytest.raises(RuntimeError, match="Test error"):
        list(llm._prepare_input_and_invoke_stream(prompt="Hello"))

    mock_body.close.assert_called_once()


@pytest.mark.asyncio
async def test_astream_closes_response_body(
    anthropic_streaming_response_with_close,
) -> None:
    mock_client = MagicMock()
    mock_client.invoke_model_with_response_stream.return_value = (
        anthropic_streaming_response_with_close
    )

    llm = BedrockLLM(
        client=mock_client,
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        region="us-west-2",
    )

    chunks = []
    async for chunk in llm._aprepare_input_and_invoke_stream(prompt="Hello"):
        chunks.append(chunk)

    assert len(chunks) > 0
    anthropic_streaming_response_with_close["body"].close.assert_called_once()


@pytest.mark.asyncio
async def test_astream_closes_response_body_on_exception() -> None:
    mock_client = MagicMock()
    mock_body = MagicMock()
    mock_client.invoke_model_with_response_stream.return_value = {"body": mock_body}

    llm = BedrockLLM(
        client=mock_client,
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        region="us-west-2",
    )

    async def mock_adapter_stream(*args, **kwargs):
        yield AIMessageChunk("test")
        raise RuntimeError("Test error")

    with patch(
        "langchain_aws.llms.bedrock.LLMInputOutputAdapter.aprepare_output_stream",
        side_effect=lambda *args, **kwargs: mock_adapter_stream(*args, **kwargs),
    ):
        with pytest.raises(RuntimeError, match="Test error"):
            async for _ in llm._aprepare_input_and_invoke_stream(prompt="Hello"):
                pass

    mock_body.close.assert_called_once()
